//---------------------------------------------------
// module  : EdgeSpecif.h                         
// auteur  : Max Mignotte                            
// date    :                                              
// langage : C++                                      
// labo    :                                  
//---------------------------------------------------

#ifndef EDGE_SPECIF_H
#define EDGE_SPECIF_H

//------------------------------------------------
// FICHIERS INCLUS -------------------------------
//------------------------------------------------
#include "FunctionEdgeSpecif.h"
#include "Chrono.h"

//------------------------------------------------
// DEFINITIONS -----------------------------------
//------------------------------------------------
#define NCHAR 200
#define TROIS  3
#define QUATRE 4
 

//>PARAMETERS_SPEC
//======================================

#define GRADMAX  0.5            //For Uniform Law
#define NBBIN_IMG 256           //For Grad&Luminance histograms

#define NBITERMAX_SPEC 6        //Iterations Specif
#define NBITERMAX_GRAD 20       //Iterations Gradient
#define NBITERMAX_METR 100      //Iterations Stochastic 

#define STEP_GRADIENT 0.5       //Step Gradient
    
#define TEMP_INIT 0.00003       //Temp.Init Stochastic
#define TEMP_FINAL 0.0000000005 //Temp.Final Stochastic
#define RADIUS_EXPLOR 0.04      //Radius Exploration Stochastic

#define DIST_BHATTA_MAX 0.1     //Stopping Criterion 
#define ENERGY_MIN 0.1          //Stopping Criterion 


//------------------------------------------------
// CLASSE EDGE_SPECIF ----------------------------
//------------------------------------------------
class EdgeSpecif
{
  //--Donnees--
  //-----------
  public:

  //>MtxResult
  float***  Imag1;
  float***  Imag2;
  float***  Imag3;

  //>Gradient
  float**   Img1;
  float**   Img2;
  float**   Img3;
  float**   Img4;
  float**   Img5;
  float**   Img6;
  float**   Img7;
  float**   Img8;
  float**   Img9;

  //>SPEC
  float** Hist1;
  float** Hist2;
  float** Hist3;

   //>Utile
  char Name_Img[NCHAR];
  char Name_Img1[NCHAR];
  char Name_Img2[NCHAR];
  char Extension[NCHAR];  
  char InfoProcess[100*NCHAR];
  
   //>Flag
  int  flag_save;
  int  flag_visu;
  int  Flag_SPEC;

  
  //>Parametres <IMG>
  int width,length;
  int width2,length2;

  //>Parametre <ALGO>
  float Alpha;      //Method
  int   LawNumber;  //Law or Neighborhodd Order

  int   SzW;        //TailleWindow
  float StGr;       //StepGradient
  int   NbIter;     //NbIterations

  //>Mesures Finales <ALGO>
  float StGrInit;
  int   TimeSec;
  int   NbIterFinal;

  float NrjImgInit,NrjImg; 
  float BhatBef,BhatAft;
  
  
  public:
  EdgeSpecif(int,int,int,int); 
  ~EdgeSpecif();

  //>Method
  void Specif();


 //=================================
 // SPEC :: File>f_EdgeSpecif2.cc
 //=================================

 //>Histo
 float SPEC_HistoModels(float,float,int);
 void  SPEC_ComputeGradHisto(float*,int,float);
 void  SPEC_ComputeGradHisto(float**,float*,int,int,int);
 void  SPEC_ComputeGradHisto(float**,int,float*,int,int);
 void  SPEC_ComputeHisto(float**,float*,int,int);
 void  SPEC_ViewHisto(float*,char*,int);
 void  SPEC_CopyHistoInMatrixForVisu(float*,float**,int);
 void  SPEC_CopyHistoInMatrixForVisu(float*,float**,int,float);

 //>Specif:::HistoLuminance[Coltuc]
 void  SPEC_SpecifLumHistoFromImg(float**,float**,int,int,float**,int,int);
 void SPEC_SpecifHistoFromDist(float**,float**,int,int,float*); 
 
 //>Specif:::HistoCouleur
 void SPEC_SpecifColHistoFromImg(float***,float***,int,int,float***,int,int); 
 void SPEC_SpecifColHistoFromImg2(float***,float***,int,int,float***,int,int); 
 void SPEC_SpecifHistoFromImgNP(float***,float***,int,int,float***,int,int);

 //>Specif:::HistoGradLuminance
 void SPEC_SpecifGradHistoFromMod(float**,int,int,int,int,float**,int,int,int);
 void SPEC_SpecifGradHistoFromImg(float**,int,int,int,int,float**,float**,int,int,int);

 //Specif:::Optimisation
 void SPEC_ComputeNonLocalGraphWindow(float**,int,int,float**,int,int,float*,int);
 void SPEC_ComputeNonLocalGraphWindow(float**,int,int,float**,int,int,float**,int);
 void SPEC_ConjGradDesc(float**,float**,int,int,float**,int,int);
 void SPEC_DeriveNrj(float**,float**,int,int,float**,int);
 float SPEC_ComputeNrj(float**,int,int,float**,int);
 void SPEC_GivePosPixel(int,int,int,int*,int*,int,int,int);
 void SPEC_SpecHistoGradOnModel(float**,int,int,int,float*,int);
 void SPEC_SpecHistoGradOnModel(float**,int,int,int,float**,int);

 //>Color & Luminance Treatment
 void SPEC_ConvertImgRGBToLum(float***,float**,int,int);
 void SPEC_ConvertImgRGBToLum_(float***,float**,int,int);
 void SPEC_ReFindColor(float**,float***,float***,int,int);
 void SPEC_BuiltColorImg(float***,float**,int,int,float***,float);
 void SPEC_RecalSameMoyLum(float***,float**,int,int);

 //>Distance
 float SPEC_ComputeBhattaDist(float*,float*,int);

 //>Info
 void SPEC_FillInfoProcess();

 //>Recuit Simule 
 float SPEC_ComputeLocNrj(float**,int,int,float,int,int,float**,int,int);
 float SPEC_DiffMat(float**,float**,int,int,float);
 void  SPEC_SimulatedAnnealing(float**,float**,int,int,float**,int,int,float,float);
};

#endif 
