//------------------------------------------------------
// module  : FunctionEdgeSpecif.cc
// auteur  : Mignotte Max
// date    :
// version : 1.0
// langage : C++
// labo    : DIRO
// note    :
//------------------------------------------------------
// quelques fonctions de reservations memoires utiles 

//------------------------------------------------
// FICHIERS INCLUS -------------------------------
//------------------------------------------------
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <iostream.h>
#include <new.h>

#include "FunctionEdgeSpecif.h"


//-------------------------//
//-- Matrice de Flottant --//
//-------------------------//
//---------------------------------------------------------
//  alloue de la memoire pour une matrice 1d de float
//----------------------------------------------------------
float* fmatrix_allocate_1d(int hsize)
 {
  float* matrix;
  matrix=new float[hsize]; return matrix; }

//----------------------------------------------------------
//  alloue de la memoire pour une matrice 2d de float
//----------------------------------------------------------
float** fmatrix_allocate_2d(int vsize,int hsize)
 {
  float** matrix;
  float *imptr;

  matrix=new float*[vsize];
  imptr=new  float[(hsize)*(vsize)];
  for(int i=0;i<vsize;i++,imptr+=hsize) matrix[i]=imptr;
  return matrix;
 }

//----------------------------------------------------------
// alloue de la memoire pour une matrice 3d de float
//----------------------------------------------------------
float*** fmatrix_allocate_3d(int dsize,int vsize,int hsize)
 {
  float*** matrix;

  matrix=new float**[dsize];

  for(int i=0;i<dsize;i++)
    matrix[i]=fmatrix_allocate_2d(vsize,hsize);
  return matrix;
 }


//-----------------------//
//-- Matrice d'entiers --//
//-----------------------//
//---------------------------------------------------------
//  alloue de la memoire pour une matrice 1d de int
//----------------------------------------------------------
int* imatrix_allocate_1d(int hsize)
 {
  int* matrix;
  matrix=new int[hsize]; return matrix; }

//----------------------------------------------------------
//  alloue de la memoire pour une matrice 2d de int
//----------------------------------------------------------
int** imatrix_allocate_2d(int vsize,int hsize)
 {
  int** matrix;
  int*  imptr;

  matrix=new int*[vsize];
  imptr=new  int[(hsize)*(vsize)];
  for(int i=0;i<vsize;i++,imptr+=hsize) matrix[i]=imptr;
  return matrix;
 }

//---------------------------//
//--- Liberation memoire ----//
//---------------------------//
//----------------------------------------------------------
// libere la memoire de la matrice 1d de float
//----------------------------------------------------------
void free_fmatrix_1d(float* pmat)
{ delete[] pmat; }

//----------------------------------------------------------
// libere la memoire de la matrice 2d de float
//----------------------------------------------------------
void free_fmatrix_2d(float** pmat)
{ delete[] (pmat[0]);
  delete[] pmat;}

//----------------------------------------------------------
// libere la memoire de la matrice 3d de float
//----------------------------------------------------------
void free_fmatrix_3d(float*** pmat,int dsize)
{
 for(int i=0;i<dsize;i++)
  {
   delete[] (pmat[i][0]);
   delete[] (pmat[i]);
   }
 delete[] (pmat);
}

//----------------------------------------------------------
// libere la memoire de la matrice 1d de int
//----------------------------------------------------------
void free_imatrix_1d(int* pmat)
{ delete[] pmat; }

//----------------------------------------------------------
// libere la memoire de la matrice 2d de int
//----------------------------------------------------------
void free_imatrix_2d(int** pmat)
{ delete[] (pmat[0]);
  delete[] pmat;}


//--------------------//
//--------------------//
//-- Matrix Gestion --//
//--------------------//
//----------------------------------------------------------
// copie une matrice dans une autre
//----------------------------------------------------------
void CopyMat(float** mat1,float** mat2,int lgth,int wdth)
{
 int i,j;

  for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)
  mat1[i][j]=mat2[i][j]; 
}

//---------------------------------------------------
// Recal                                  
//---------------------------------------------------
void Recal(float** mat,int lgth,int wdth)
{
 int i,j;
 float max,min;

 //Initialisation
 max=-1000000.0;
 min=100000000;

 //Recherche du min
  for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)
    if (mat[i][j]<min) min=mat[i][j];

 //plus min
   for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)
    mat[i][j]-=min;

 //Recherche du max
  for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
    if (mat[i][j]>max) max=mat[i][j];

 //Recalibre la matrice
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)
   if (max) mat[i][j]*=(GREY_LEVEL/max);      
}

//---------------------------------------------------
// RecalI                                  
//---------------------------------------------------
void RecalI(float** mat,int lgth,int wdth,float min,float max)
{
 int i,j; 
 float scale;
 scale=max-min;

 //Recal [0-255]
 Recal(mat,lgth,wdth);
 
 //Recal [min-max]
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) mat[i][j]*=scale/255.0;
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) mat[i][j]+=min; 
}

//---------------------------------------------------
// MultMatrix                               
//---------------------------------------------------
void MultMatrix(float** mat,int lgth,int wdth,float mult)
{
 for(int i=0;i<lgth;i++) for(int j=0;j<wdth;j++) mat[i][j]*=mult; 
}

//----------------------------------------------------------
// ComputeMean
//----------------------------------------------------------
float ComputeMean(float** mat,int lgth,int wdth)
{
 int i,j;
 float Moy;

 Moy=0.0;
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) Moy+=mat[i][j]; 
 Moy/=(lgth*wdth);

 return Moy;
}

//----------------------------------------------------------
// ComputeDiff
//----------------------------------------------------------
float ComputeDiff(float** mat1,float** mat2,int lgth,int wdth)
{
 int i,j;
 float Diff;

 Diff=0.0;
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) Diff+=fabs(mat1[i][j]-mat2[i][j]);
 Diff/=(lgth*wdth);

 return Diff;
}

//----------------------------------------------------------
// Stat           
//----------------------------------------------------------     
void Stat(float** Img,int lgth,int wdth)
{
 int i,j;
 float min,max,moy; 

 //Stats
 min=10000000.0;
 max=-10000000.0;
 moy=0.0;
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)
    { moy+=Img[i][j];
      if(min>Img[i][j]) min=Img[i][j]; 
      if(max<Img[i][j]) max=Img[i][j];  }
  moy/=(lgth*wdth);
  printf("\n Stat::[min=%.1f | max=%.1f | Moy=%.1f]",min,max,moy);
  fflush(stdout);
}

//----------------------//
//----------------------//
//-- FILE -PGM/PPM------//
//----------------------//
//----------------------------------------------------------
// Get Length and Width
//----------------------------------------------------------
void GetLengthWidth(char* path,int* length,int* width)
{
 unsigned char var;
 int   temp;
 char* tempc;
 char stringTmp1[100];
 char stringTmp2[100];
 int ta1,ta2;
 FILE *fic;

  //ouverture du fichier
  fic=fopen(path,"r");
  if (fic==NULL)
    { printf("\n- Grave erreur a l'ouverture de %s -\n",path);
      exit(-1); }

  //recuperation de l'entete
  tempc=fgets(stringTmp1,100,fic);
  for(;;) { temp=fread(&var,1,1,fic); if (var==35) tempc=fgets(stringTmp2,100,fic);
            else break; }
  fseek(fic,-1,SEEK_CUR);
  temp=fscanf(fic,"%d %d",&ta1,&ta2);
   
  //enregistrement
  (*length)=ta2; 
  (*width)=ta1;

  //fermeture du fichier
  fclose(fic);
}

//----------------------------------------------------------
// Chargement de l'image de nom <name> (en ppm)             
//----------------------------------------------------------
void load_image_ppm(char* path,float*** img,int lgth,int wdth)
 {
  int i,j;
  int temp;
  char* tempc;
  unsigned char varb,varv,varr;
  char Buff[600],stringTmp1[600],stringTmp2[600];
  int ta1,ta2,ta3;
  FILE *fic;

  //enregistrement du chemin;
   strcpy(Buff,path);
   if (strstr(Buff,".ppm")==NULL) strcat(Buff,".ppm");

   //ouverture du fichier
  fic=fopen(Buff,"r");
  if (fic==NULL) 
   { printf("\n ->> Grave erreur a l'ouverture de %s -\n",Buff);
     exit(-1); }
   printf("\n Ouverture de -> %s au format ppm",Buff);
   fflush(stdout);
  
  //recuperation de l'entete
  tempc=fgets(stringTmp1,100,fic); //P6
  tempc=fgets(stringTmp2,600,fic); //Comment

  for(;;)
    { if (((int)stringTmp2[0]==10)||((int)stringTmp2[0]==35))
      tempc=fgets(stringTmp2,600,fic); 

      else break; }

  temp=sscanf(tempc," %d %d",&ta1,&ta2);
  temp=fscanf(fic," %d",&ta3);

  //affichage de l'entete
  //printf("\n [%s::%d::%d::%d]",stringTmp1,ta1,ta2,ta3);
  printf("  [%d::%d::%d]",ta1,ta2,ta3);
  fflush(stdout);
   
  //chargement dans la matrice
  for(i=0;i<lgth;i++) for(j=0;j<(wdth*3);j+=3)  
    { temp=fread(&varr,1,1,fic);
      temp=fread(&varv,1,1,fic);
      temp=fread(&varb,1,1,fic);
     
      img[2][i][j/3]=varr;        
      img[0][i][j/3]=varv;        
      img[1][i][j/3]=varb; }      

  //fermeture du fichier
  fclose(fic);
 }

//----------------------------------------------------------
// Sauvegarde de l'image de nom <name> au format ppm        
//----------------------------------------------------------
void SaveImagePpm(char* Path_out,char* Name,char* Com,float*** matrvb,int lgth,int wdth)
 {
   int i,j,k;
  char buff[200];
  FILE* fuser;

  //extension
  strcpy(buff,Path_out);
  strcat(buff,Name);
  strcat(buff,".ppm");

  //ouverture fichier
  fuser=fopen(buff,"w");
    if (fuser==NULL) 
        { printf(" probleme dans la sauvegarde de %s",buff); 
          exit(-1); }

  //affichage
  printf("\n Sauvegarde dans -> %s au format %s",buff,"ppm");
  fflush(stdout);

  //sauvegarde de l'entete
  fprintf(fuser,"P6");
  fprintf(fuser,"\n# %s \n",Com);
  fprintf(fuser,"%d %d",wdth,lgth);
  fprintf(fuser,"\n255\n");

  for(k=0;k<3;k++) for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)
    { matrvb[k][i][j]=(int)(matrvb[k][i][j]+0.49);
      if (matrvb[k][i][j]<0)   matrvb[k][i][j]=0;
      if (matrvb[k][i][j]>255) matrvb[k][i][j]=255; }      

  //enregistrement
  for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
    { fprintf(fuser,"%c",(char)matrvb[0][i][j]);
      fprintf(fuser,"%c",(char)matrvb[1][i][j]);
      fprintf(fuser,"%c",(char)matrvb[2][i][j]); }
       
  //fermeture fichier
   fclose(fuser); 
 }

//--------------------------------
//--------------------------------
//----CONJUGATE GRADIENT----------
//--------------------------------
//------------------------------------------------------------------------
// ComputeBeta  
// <Grad_J(x[n]-Grad_J(x[n-1]),Grad_J(x[n]>/<Grad_J(x[n-1]),Grad_J(x[n-1])>  
//------------------------------------------------------------------------  
float ComputeBeta(float** ImGrd,float** ImGrdp,int lgth, int wdth)
{
 int i,j;
 float Num,Denum;
 float Res=0.0;

 //AllocationMemoire
 float** ImTmp=fmatrix_allocate_2d(lgth,wdth);

 // ImRes <-- Grad_J(x[n])-Grad_J(x[n-1])
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) ImTmp[i][j]=ImGrd[i][j]-ImGrdp[i][j];

 // Numerateur & Denumerateur
 Num=0.0;
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) Num+=ImTmp[i][j]*ImGrd[i][j];
 Denum=0.0;
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) Denum+=CARRE(ImGrdp[i][j]);

 //Liberation Memoire
 if (ImTmp) free_fmatrix_2d(ImTmp);

 //Retour
 if (Denum) Res=Num/Denum;
 return Res;
}


//----------------------------------------------------------
// AddMatrix   <matres=mata+fact*matb>               
//----------------------------------------------------------     
void AddMatrix(float** matout,float** matin1,float fact,float** matin2,int lgth,int wdth)
{
 int i,j;

 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
     matout[i][j]=matin1[i][j]+(fact*matin2[i][j]);
}


//----------------------------------------------------------
// MultMatrix   <matout=matin1*matin2>               
//----------------------------------------------------------     
void MultMatrix(float** matout,float Beta,float** matin,int lgth,int wdth)
{
 int i,j;

 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) matout[i][j]=Beta*matin[i][j];
}

//----------------------------------------------------------
// InitMatrix
//----------------------------------------------------------     
void InitMatrix(float** mat,int lgth,int wdth)
{
 int i,j;

 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) mat[i][j]=0.0;
}




//--------------------------------
//--------------------------------
//---GESTION COULEUR--------------
//--------------------------------
//--------------------------------
//---LAB>RGB----------------------
//--------------------------------
//----------------------------------------------------------*/
// LABToRGB                    
//----------------------------------------------------------*/
void LABToRGB(float l,float a,float b,float* r,float* g,float* bl)
{
 float x,y,z;

 LABToXYZ(l,a,b,&x,&y,&z);
 XYZToRGB(x,y,z,r,g,bl);
}

//----------------------------------------------------------
// LABToXYZ 
//----------------------------------------------------------
void LABToXYZ(float l,float a,float b,float* x,float* y,float* z)
{
 (*y)=(l+16.0)/116.0;
 (*x)=a/500.0+(*y);
 (*z)=(*y)-b/200.0;

 if (powf((*y),3)>0.008856)  (*y)=powf((*y),3);
 else                        (*y)=((*y)-16.0/116.0)/7.787;
 if (powf((*x),3)>0.008856)  (*x)=powf((*x),3);
 else                        (*x)=((*x)-16.0/116.0)/7.787;
 if (pow((*z),3)>0.008856)   (*z)=powf((*z),3); 
 else                        (*z)=((*z)-16.0/116.0)/7.787;

  (*x)=95.047*(*x);     
  (*y)=100.000*(*y);     
  (*z)=108.883*(*z);     
}

//----------------------------------------------------------
// XYZToRGB 
//----------------------------------------------------------
void XYZToRGB(float x,float y,float z,float* r,float* g,float* b)
{
 x=x/100.0;   //X from 0 to  95.047
 y=y/100.0;   //Y from 0 to 100.000
 z=z/100.0;   //Z from 0 to 108.883
  
 (*r)=x*3.2406+y*(-1.5372)+z*(-0.4986);
 (*g)=x*(-0.9689)+y*(1.8758)+z*0.0415;
 (*b)=x*0.0557+y*(-0.2040)+z*1.0570;

 if ((*r)>0.0031308) (*r)=1.055*(powf((*r),(1.0/2.4)))-0.055;
 else                (*r)=12.92*(*r);
 if ((*g)>0.0031308) (*g)=1.055*(powf((*g),(1.0/2.4)))-0.055;
 else                (*g)=12.92*(*g);
 if ((*b)>0.0031308) (*b)=1.055*(powf((*b),(1.0/2.4)))-0.055;
 else                (*b)=12.92*(*b);

 (*r)=(*r)*255.0;
 (*g)=(*g)*255.0;
 (*b)=(*b)*255.0;
}

//--------------------------------
//---RGB>LAB----------------------
//--------------------------------
//----------------------------------------------------------*/
// RGBToLAB                     
//----------------------------------------------------------*/
void RGBToLAB(float r,float g,float bl,float* l,float* a,float* b)
{
 float x,y,z;

 RGBToXYZ(r,g,bl,&x,&y,&z);
 XYZToLAB(x,y,z,l,a,b);
}

//----------------------------------------------------------*/
// RGBToXYZ                    
//----------------------------------------------------------*/
void RGBToXYZ(float r,float g,float b,float* x,float* y,float* z)
{
  r=(r/255.0);         //R from 0 to 255
  g=(g/255.0);         //G from 0 to 255
  b=(b/255.0);         //B from 0 to 255

  if (r>0.04045) r=powf(((r+0.055)/1.055),2.4);
  else           r=r/12.92;
  if (g>0.04045) g=powf(((g+0.055)/1.055),2.4);
  else           g=g/12.92;
  if (b>0.04045) b=powf(((b+0.055)/1.055),2.4);
  else           b=b/12.92;

  r=r*100;
  g=g*100;
  b=b*100;

  (*x)=r*0.4124 + g*0.3576 + b*0.1805;
  (*y)=r*0.2126 + g*0.7152 + b*0.0722;
  (*z)=r*0.0193 + g*0.1192 + b*0.9505;
}

//----------------------------------------------------------*/
// XYZToLAB                     
//----------------------------------------------------------*/
void XYZToLAB(float x,float y,float z,float* l,float* a,float* b)
{
  x=x/95.047;           
  y=y/100.000;          
  z=z/108.883;       

  if (x>0.008856) x=powf(x,1.0/3.0);
  else            x=(7.787*x)+(16.0/116.0);
  if (y>0.008856) y=powf(y,1.0/3.0);
  else            y=(7.787*y)+(16.0/116.0);
  if (z>0.008856) z=powf(z,1.0/3.0); 
  else            z=(7.787*z)+(16.0/116.0);

  (*l)=(116.0*y)-16.0;
  (*a)=500.0*(x-y);
  (*b)=200.0*(y-z);
}



//----------------------------------------------------------
// Convert_LAB_TO_RGB                 
//----------------------------------------------------------
void Convert_LAB_TO_RGB(float*** mat,int lgth,int wdth,float*** matout)
{
 int i,j,l;
 int cptU,cptD,cptUD;
 float red,green,blue; 
 float l_,a_,b_;
 float pourcU,pourcD,pourcUD;

 //Presentation
 printf("\n\n >>> ConvertPourc_LAB_TO_RGB\n");

 //Init Recalage LAB 
  RecalI(mat[0],lgth,wdth,0.0,100.0);
  RecalI(mat[1],lgth,wdth,-60.0,80.0);
  RecalI(mat[2],lgth,wdth,-60.0,80.0);

  Stat(mat[0],lgth,wdth);
  Stat(mat[1],lgth,wdth);
  Stat(mat[2],lgth,wdth);
  printf("\n ");

 //AllocationMemoire
 float*** mattmp=fmatrix_allocate_3d(TROIS,lgth,wdth);
 float*** mattmp2=fmatrix_allocate_3d(TROIS,lgth,wdth);

 //Final Stretching
 for(l=0;l<TROIS;l++) for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) mattmp[l][i][j]=mat[l][i][j];
 RecalI(mattmp[0],lgth,wdth,0.0,100.0);

 //LAB Final Stretching > mattmp2
 for(l=0;l<TROIS;l++) for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) mattmp2[l][i][j]=mattmp[l][i][j];
 
 //LAB>RGB
 cptD=0;
 cptU=0;
 cptUD=0;
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
    { LABToRGB(mattmp[0][i][j],mattmp[1][i][j],mattmp[2][i][j],&red,&green,&blue);
      if ((red>255)||(green>255)||(blue>255))  cptU++; 
      if ((red<0)||(green<0)||(blue<0))        cptD++;
      if ((red>255)||(green>255)||(blue>255)||((red<0)||(green<0)||(blue<0))) cptUD++;
     
      if (red>255)   red=255;
      if (green>255) green=255;
      if (blue>255)  blue=255;
      if (red<0)     red=0;
      if (green<0)   green=0;
      if (blue<0)    blue=0;

      matout[0][i][j]=(int)red;
      matout[1][i][j]=(int)green;
      matout[2][i][j]=(int)blue;  }
 
  pourcU=(float)100.0*cptU/(lgth*wdth);
  pourcD=(float)100.0*cptD/(lgth*wdth);
  pourcUD=(float)100.0*cptUD/(lgth*wdth);
  printf(" \n [pcU=%.1f][pcD=%.1f][pcUD=%.1f]",pourcU,pourcD,pourcUD);
 
 //Re-Verif
  cptUD=0;
  for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
    { RGBToLAB(matout[0][i][j],matout[1][i][j],matout[2][i][j],&l_,&a_,&b_);
      if ((fabs(l_-mattmp2[0][i][j])>0.5)||(fabs(a_-mattmp2[1][i][j])>0.5)||(fabs(b_-mattmp2[2][i][j])>0.5))
      cptUD++; }
  pourcUD=(float)100.0*cptUD/(lgth*wdth);
  printf(" >> [pcUD=%.2f]",pourcUD);
   
 //Liberation Memoire
 if (mattmp)   free_fmatrix_3d(mattmp,TROIS);
 if (mattmp2)  free_fmatrix_3d(mattmp2,TROIS);
}

//----------------------------------------------------------
// Convert RGBToLAB
//----------------------------------------------------------
void ConvertImgRGBToLAB(float*** RGB,float*** LAB,int lgth,int wdth)
{
 int i,j;

 //Boucle
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)
 RGBToLAB(RGB[0][i][j],RGB[1][i][j],RGB[2][i][j],&(LAB[0][i][j]),&(LAB[1][i][j]),&(LAB[2][i][j]));
}

//----------------------------------------------------------
// Convert RGBToLAB
//----------------------------------------------------------
void ConvertImgLABToRGB(float*** LAB,float*** RGB,int lgth,int wdth)
{
 int i,j;

 //Boucle
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)
 LABToRGB(LAB[0][i][j],LAB[1][i][j],LAB[2][i][j],&(RGB[0][i][j]),&(RGB[1][i][j]),&(RGB[2][i][j]));
}

//-------------//
//-------------//
// SORT -------//
//-------------//
//----------------------------------------------------------
// retourne un nombre aleatoire entre zero et un
//----------------------------------------------------------
float randomize(void)
{ return ((float)rand()/RAND_MAX); }

//----------------------------------------------------------------
//  quickSort
//----------------------------------------------------------------
void swap(float* a,float* b)
{
  float t=*a; *a=*b; *b=t;
}

void sort(float* arr,int beg,int end)
{
  if (end>beg+1)
  {
    float piv=arr[beg];
    int l=beg+1,r=end;
    while (l<r)
    {
      if (arr[l]<=piv) l++;
      else swap(&arr[l],&arr[--r]);
    }
    swap(&arr[--l],&arr[beg]);
    sort(arr,beg,l);
    sort(arr,r,end);
  }
}

//----------------------------------------------------------------
//  quickSort
//  ---------     Arr[l][c]   :: Arr[l][0] a classer
//                            :: l \in [beg::end]
//                            :: c \in [0::Nb]
//----------------------------------------------------------------
void sort(float** arr,int beg,int end,int Nb)
{
 int k;

 //>Sort
 if (end>beg+1)
  {
    float piv=arr[beg][0];
    int l=beg+1,r=end;
    while (l<r)
    {
      if (arr[l][0]<=piv) l++;
      else { r--; for(k=0;k<Nb;k++) swap(&arr[l][k],&arr[r][k]); }
    }
    l--; for(k=0;k<Nb;k++) swap(&arr[l][k],&arr[beg][k]);
    sort(arr,beg,l,Nb);
    sort(arr,r,end,Nb);
  }
}

//-------------//
//-------------//
// MISC--------//
//-------------//

//----------------------------------------------------------
// FuncGauss
//----------------------------------------------------------
float funcgauss(float pix,float moy,float var)
{
return (float)((1/(sqrt(2*PI*var)))*exp(-1*(CARRE((float)pix-moy))/(2*var)));
}

//----------------------------------------------------------
// IntMat Mat [0.0::1.0]
//----------------------------------------------------------
void IntMat(float** Mat,int lgth,int wdth)
{
  int i,j;

  for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) Mat[i][j]=Mat[i][j]*256.0;
  for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) Mat[i][j]=(int)(Mat[i][j]);
  for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) Mat[i][j]=Mat[i][j]/256.0;;
}
