//---------------------------------------------------
// module  : f_EdgeSpecif2.cc                       
// auteur  : Max Mignotte                            
// date    :                                              
// langage : C++                                      
// labo    :                                  
//---------------------------------------------------
//
//________________
//_____ SPEC _____
//________________
//________________
//-----------------------------------------------
// FICHIERS INCLUS ------------------------------
//-----------------------------------------------
#include "EdgeSpecif.h"

//======================================
//======================================
//  HISTOGRAMME  [CALCUL/VISU/MODELS]
//======================================
//======================================
//----------------------------------------------------------
// SPEC >> SPEC_HistoModels
//----------------------------------------------------------
//
//  nb=0 >  Uniform
//     1 >  SemiGaussienne
//     2 >  Triangle
//     3 >  Shifted Gaussien
//     4 >  Decreasing Exponential
//
//     5 > 1 EDge Class
//     6 > 2 Edge Class
//     7 > 3 Edge Class
//     8 > 4 Edge Class
//
//---------------------------------------------------------- 
float EdgeSpecif::SPEC_HistoModels(float k,float par,int LawNb)
{
 float tmp;

  //Uniform [-]
  if (LawNb==0)
     { if (k<GRADMAX) return 1.0;
       else return 0.0; }

  //SemiGaussienne [Ok <1>]
  if (LawNb==1)
     { tmp=(k*NBBIN_IMG);
       tmp=exp(-CARRE(1.5*par*tmp));
	return tmp; }
 
 //Triangle [Ok <2>]
 if (LawNb==2)
   {  tmp=k*NBBIN_IMG;
      tmp=1.0-(0.25*par*tmp);
      if (tmp>0) return tmp;
      else return 0.0; }

 //Gaussienne Decale [Ok <3>]
 if (LawNb==3)
    { tmp=funcgauss(k,0.1,0.0001); 
      return tmp; }

 //Exponentiel Decroissante [Ok <4>]
 if (LawNb==4)
   { tmp=exp(-8*k); 
      return tmp; }

 //===============================
 //==EDGE-CLASSES=================
 //===============================

 //Gaussienne <1 classe> [Ok <A>]
 if (LawNb==5)
   { tmp=funcgauss(k,0.0,0.0001); 
     return tmp; }

 //Gaussienne <2 classes> [Ok <B>]
 if (LawNb==6)
   { tmp=funcgauss(k,0.0,0.0001)+0.025*funcgauss(k,0.20,0.0001);
     return tmp; }

  //Gaussienne <3 classes> []
 if (LawNb==7)
   { tmp=funcgauss(k,0.0,0.0001)+0.1*funcgauss(k,0.10,0.0001)+0.1*funcgauss(k,0.20,0.0001); 
     return tmp; }

  ///Gaussienne <4 classes> []
 if (LawNb==8)
   { tmp=funcgauss(k,0.0,0.0001)+0.05*funcgauss(k,0.10,0.0001)+0.05*funcgauss(k,0.15,0.0001)+0.05*funcgauss(k,0.20,0.0001); 
     return tmp; }

  //==End_EDGE_CLASSES=============
 
 return 0.0;
}

//----------------------------------------------------------
// SPEC >> SPEC_ComputeGradHisto
//----------------------------------------------------------  
//  
// order [0:Unif   1:Trgl   2:SemCirc   3:3Dirac]
//
//----------------------------------------------------------     
void EdgeSpecif::SPEC_ComputeGradHisto(float* HistGrad,int LawNb,float par)
{
 int k;
 float tmp;

 for(k=0;k<NBBIN_IMG;k++) HistGrad[k]=SPEC_HistoModels((float)k/NBBIN_IMG,par,LawNb);
 for(tmp=0.0,k=0;k<NBBIN_IMG;k++) tmp+=HistGrad[k];
 for(k=0;k<NBBIN_IMG;k++) if (tmp) HistGrad[k]/=tmp;
}

//----------------------------------------------------------
// SPEC >> SPEC_ComputeGradHisto
//----------------------------------------------------------  
// Nota: Img in [0.0::1.0]
//       HistoGrad[NBBIN_IMG]
//  
//----------------------------------------------------------     
void EdgeSpecif::SPEC_ComputeGradHisto(float** Img,float* HistoGrad,int Nghb,int lgth,int wdth)
{
 int i,j,k;
 float tmp;
 float moy,var;

 //Presentation  
 printf("\n  >> Calcul Histo Gradient");
 
 //>Compute GradHisto
 for(i=0;i<NBBIN_IMG;i++) HistoGrad[i]=0.0;
 for(k=1;k<=Nghb;k++)
   { for(i=Nghb;i<(lgth-Nghb);i++) for(j=Nghb;j<(wdth-Nghb);j++)
      { tmp=NBBIN_IMG*fabs(Img[i+Nghb][j]-Img[i][j]);
        HistoGrad[(int)(tmp)]++;
        tmp=NBBIN_IMG*fabs(Img[i][j+Nghb]-Img[i][j]);
        HistoGrad[(int)(tmp)]++;
        tmp=NBBIN_IMG*fabs(Img[i+Nghb][j+Nghb]-Img[i][j]);
        HistoGrad[(int)(tmp)]++; } }

 //>Normalisation Histo
 for(tmp=0.0,k=0;k<NBBIN_IMG;k++) tmp+=HistoGrad[k];
 for(k=0;k<NBBIN_IMG;k++) if (tmp) HistoGrad[k]/=tmp;

 //>Calcul Moy/Var
 for(moy=0.0,k=0;k<NBBIN_IMG;k++) moy+=((float)k/NBBIN_IMG)*HistoGrad[k];
 for(var=0.0,k=0;k<NBBIN_IMG;k++) var+=CARRE(((float)k/NBBIN_IMG)-moy)*HistoGrad[k];
 printf(" (Var[%.2f])",var*100);
 fflush(stdout);

 //>DEBUG
 if (0) for(k=0;k<NBBIN_IMG;k++) printf("[%d::%.3f]",k,HistoGrad[k]);
 if (0) for(k=0;k<NBBIN_IMG;k++) printf("\n %.3f",HistoGrad[k]);
}

//----------------------------------------------------------
// SPEC >> SPEC_ComputeGradHisto
//----------------------------------------------------------  
// Nota: Img in [0.0::1.0]
//       HistoGrad[NBBIN_IMG]
//  
//----------------------------------------------------------     
void EdgeSpecif::SPEC_ComputeGradHisto(float** Img,int order,float* HistoGrad,int lgth,int wdth)
{
 int i,j,k;
 float tmp;
 float moy,var;

 //Presentation  
 printf("\n  >>> Calcul Histo Gradient [order=%d]",order);
 fflush(stdout);
 
 //>Compute GradHisto
 for(i=0;i<NBBIN_IMG;i++) HistoGrad[i]=0.0;

 for(i=order;i<(lgth-order);i++) for(j=order;j<(wdth-order);j++)
      { tmp=NBBIN_IMG*fabs(Img[i+order][j]-Img[i][j]);
        HistoGrad[(int)(tmp)]++;
        tmp=NBBIN_IMG*fabs(Img[i][j+order]-Img[i][j]);
        HistoGrad[(int)(tmp)]++;
        tmp=NBBIN_IMG*fabs(Img[i+order][j+order]-Img[i][j]);
        HistoGrad[(int)(tmp)]++; } 

 //>Normalisation Histo
 for(tmp=0.0,k=0;k<NBBIN_IMG;k++) tmp+=HistoGrad[k];
 for(k=0;k<NBBIN_IMG;k++) if (tmp) HistoGrad[k]/=tmp;

 //>Calcul Moy/Var
 for(moy=0.0,k=0;k<NBBIN_IMG;k++) moy+=((float)k/NBBIN_IMG)*HistoGrad[k];
 for(var=0.0,k=0;k<NBBIN_IMG;k++) var+=CARRE(((float)k/NBBIN_IMG)-moy)*HistoGrad[k];
 printf("\n  >>> Var[%.2f]",var*100);
 fflush(stdout);

 //>DEBUG
 if (0) for(k=0;k<NBBIN_IMG;k++) printf("[%d::%.3f]",k,HistoGrad[k]);
}

//----------------------------------------------------------
// SPEC >> SPEC_ComputeHisto
//----------------------------------------------------------  
// Nota: Img in [0.0::1.0]
//       Histo[NBBIN_IMG]
//  
//----------------------------------------------------------     
void EdgeSpecif::SPEC_ComputeHisto(float** Img,float* Histo,int lgth,int wdth)
{
 int i,j,k;
 float tmp;

 //>Presentation  
 printf("\n >> Calcul Histo Color/GreyLevel");
 
 //>Compute HistoGreyLevel
 for(k=0;k<NBBIN_IMG;k++) Histo[k]=0.0;
 
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)
    { tmp=NBBIN_IMG*Img[i][j];
      if ((int)(tmp-0.5)<NBBIN_IMG) Histo[(int)(tmp-0.5)]++; }

 //>Normalisation Histo
 for(tmp=0.0,k=0;k<NBBIN_IMG;k++) tmp+=Histo[k];
 for(k=0;k<NBBIN_IMG;k++) if (tmp) Histo[k]/=tmp;

 //>DEBUG
 if (0) for(k=0;k<NBBIN_IMG;k++) printf("[%d::%.3f]",k,Histo[k]);
}

//----------------------------------------------------------
// SPEC >> SPEC_ViewHisto
//----------------------------------------------------------  
//  
//----------------------------------------------------------     
void EdgeSpecif::SPEC_ViewHisto(float* Histo,char* string,int flag)
{
  int k;

 //Affichage
  printf("\n\n Histo %s > \n",string);
  printf("\n");
 if (!flag)    for(k=0;k<50;k++) printf("[%.3f]",Histo[k]);
 if (flag==1)  for(k=0;k<100;k++) printf("[%d::%.3f]",k,Histo[k]);
 if (flag==2)  for(k=0;k<80;k++) printf("%f %.3f\n",(float)k/NBBIN_IMG,Histo[k]);

}

//----------------------------------------------------------
// SPEC >> SPEC_CopyHistoInMatrixForVisu
//----------------------------------------------------------  
// Nota: HistoGrad[NBBIN_IMG] 
//   
//----------------------------------------------------------     
void EdgeSpecif::SPEC_CopyHistoInMatrixForVisu(float* Histo,float** Img,int sze)
{
  int i,j,k;
  float tmp,tp;
  float max,fact_norm;
   
  //>Presentation  
  printf("\n  >> Fill Img. With Histo");

  //>Blanc>Img
  for(i=0;i<sze;i++) for(j=0;j<sze;j++) Img[i][j]=GREY_LEVEL;

  //>FacteurNormalisation   
  for(max=0,k=0;k<NBBIN_IMG;k++) if (max<Histo[k])  max=Histo[k]; 
  fact_norm=(float)sze/(1.05*max);

  //>DEBUG
  if (1) printf("  [max=%f::sze=%d::Fact=%.2f]",max,sze,fact_norm);
    
  //>DessineHisto
  for(k=0;k<=NBBIN_IMG;k++)
    { tmp=(int)(Histo[k]*fact_norm);
      if (tmp>=(sze-1)) tmp=(sze-1);
      for(tp=tmp;tp>0;tp--) Img[(int)(sze-tp)][k]=GREY_LEVEL/3; }

  //>DEBUG
  if (0) for(k=0;k<NBBIN_IMG;k++) printf("[%d::%.3f]",k,Histo[k]);
}

//----------------------------------------------------------
// SPEC >> SPEC_CopyHistoInMatrixForVisu
//----------------------------------------------------------  
// Nota: HistoGrad[NBBIN_IMG] 
//   
//----------------------------------------------------------     
void EdgeSpecif::SPEC_CopyHistoInMatrixForVisu(float* Histo,float** Img,int sze,float max)
{
  int i,j,k;
  float tmp,tp;
  float fact_norm;
   
  //>Presentation  
  printf("\n  >> Fill Img. With Histo");

  //>Blanc>Img
  for(i=0;i<sze;i++) for(j=0;j<sze;j++) Img[i][j]=GREY_LEVEL;

  //>FacteurNormalisation   
  fact_norm=(float)sze/(1.05*max);

  //>DEBUG
  if (1) printf("  [max=%f::sze=%d::Fact=%.2f]",max,sze,fact_norm);
    
  //>DessineHisto
  for(k=0;k<=NBBIN_IMG;k++)
    { tmp=(int)(Histo[k]*fact_norm);
      if (tmp>=(sze-1)) tmp=(sze-1);
      for(tp=tmp;tp>0;tp--) Img[(int)(sze-tp)][k]=GREY_LEVEL/3; }

  //>DEBUG
  if (0) for(k=0;k<NBBIN_IMG;k++) printf("[%d::%.3f]",k,Histo[k]);
}


//======================================
//======================================
// SPECIFICATION HISTOGRAMME LUMINANCE
//======================================
//======================================
//----------------------------------------------------------
//  SPEC >> SpecifLumHistoFromImg
//----------------------------------------------------------
//   
// Nota : Specif HistoLum From Img [float** LumMod]
//        <Methode Coltuc>
//        Img in [0.0::1.0]
//        Histo[NBBIN_IMG]
// 
//----------------------------------------------------------     
void EdgeSpecif::SPEC_SpecifLumHistoFromImg(float** LumIn,float** LumOut,int lgth,int wdth,float** LumMod,int lgth2,int wdth2)
{
  //AllocationMemoire
  float* HistMod=fmatrix_allocate_1d(NBBIN_IMG);

  //ExactSpecif Luminance
  SPEC_ComputeHisto(LumMod,HistMod,lgth2,wdth2);
  SPEC_SpecifHistoFromDist(LumIn,LumOut,lgth,wdth,HistMod); 
  
  //Liberation Memoire
  if (HistMod)  free_fmatrix_1d(HistMod);
} 

//----------------------------------------------------------
//  SPEC >> SpecifHistoFromDist
//----------------------------------------------------------
//   
// Nota : Specif HistoLum From Distribution/Histogram
//        <Methode Coltuc>
//        Img in [0.0::1.0]
//        Histo[NBBIN_IMG]
// 
//----------------------------------------------------------     
void EdgeSpecif::SPEC_SpecifHistoFromDist(float** ImgIn,float** ImgOut,int lgth,int wdth,float* Histo)
{
 int i,j,k,l;
 int nbp,NbPixs;
 int posr,posc;
 float ng,MinDiff,PropPixs,idxmod;

 //Constantes
 NbPixs=lgth*wdth;

 //AllocationMemoire
 float** TabHist=fmatrix_allocate_2d(NbPixs,QUATRE);
 float* CumulHisto=fmatrix_allocate_1d(NBBIN_IMG);

 //>RecupValNg
 MinDiff=10000;
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)
   { nbp=(i*wdth)+j;
     TabHist[nbp][0]=ImgIn[i][j];
     TabHist[nbp][1]=i;
     TabHist[nbp][2]=j; 
     if ((ImgIn[i][j])&&(ImgIn[i][j]<MinDiff)) MinDiff=ImgIn[i][j]; }

 printf("\n  > MinDiff=%.10f",MinDiff);

 //>Add MinDiff/4 Noise
 for(k=0;k<NbPixs;k++) TabHist[k][0]+=(randomize()*(MinDiff/100.0));  
 
  //>Sort
 printf("\n  > Sort ..."); fflush(stdout);
 sort(TabHist,0,NbPixs,QUATRE);
 printf("..."); fflush(stdout);

 //>SPECIFICATION
 //============================================
 CumulHisto[0]=Histo[0];
 for(k=1;k<NBBIN_IMG;k++) CumulHisto[k]=CumulHisto[k-1]+Histo[k];
 printf("\n  > [CumHisto=%.5f]",CumulHisto[NBBIN_IMG-1]);

 for(l=0,nbp=0,k=0;k<NBBIN_IMG;k++)
   { idxmod=(float)k/(float)NBBIN_IMG;
     PropPixs=(int)((NbPixs*CumulHisto[k]));    
     for(l=nbp;(l<=(PropPixs))&&(l<NbPixs);l++) TabHist[l][3]=idxmod;    
     nbp=(int)PropPixs; }
 
 printf("\n  > [Histo::::%d>%d]",nbp,NbPixs); fflush(stdout);
 for(l=nbp;l<NbPixs;l++) if (!TabHist[l][3]) TabHist[l][3]=TabHist[l-1][3];
 //=============================================

 //>TRANSFERT>
 //===========
 for(l=0;l<NbPixs;l++)
   { ng=TabHist[l][3];
     posr=(int)TabHist[l][1];
     posc=(int)TabHist[l][2]; 
     ImgOut[posr][posc]=ng; }

 //LiberationMemoire
 if (TabHist)    free_fmatrix_2d(TabHist);
 if (CumulHisto) free_fmatrix_1d(CumulHisto);
} 

//======================================
//======================================
//  SPECIFICATION HISTOGRAMME COULEUR
//======================================
//======================================
//----------------------------------------------------------
//  SPEC >> SpecifColHistoFromImg -V1-
//----------------------------------------------------------
//   
// Nota : Specif RGB_input From Img_RGB_Target
//        <Methode Coltuc>
//        Img in [0.0::1.0]
//        Histo[NBBIN_IMG]
//        >>>Specif R,V,B
// 
//----------------------------------------------------------     
void EdgeSpecif::SPEC_SpecifColHistoFromImg(float*** ImgIn,float*** ImgOut,int lgth,int wdth,float*** ImgMod,int lgth2,int wdth2)
{
 int k;

  //AllocationMemoire
  float** HistMod=fmatrix_allocate_2d(TROIS,NBBIN_IMG);

  //ExactSpecif R/V/B
  for(k=0;k<TROIS;k++)
    { SPEC_ComputeHisto(ImgMod[k],HistMod[k],lgth2,wdth2);
      SPEC_SpecifHistoFromDist(ImgIn[k],ImgOut[k],lgth,wdth,HistMod[k]); }
  
  //Liberation Memoire
  if (HistMod)  free_fmatrix_2d(HistMod);
} 

//----------------------------------------------------------
//  SPEC >> SpecifColHistoFromImg -V2-
//----------------------------------------------------------
//   
// Nota : Specif LAB_input From Img_LAB_Target
//        <Methode Coltuc>
//        Img in [0.0::1.0]
//        Histo[NBBIN_IMG]
//        >>>Specif L,A,B
// 
//----------------------------------------------------------     
void EdgeSpecif::SPEC_SpecifColHistoFromImg2(float*** ImgIn,float*** ImgOut,int lgth,int wdth,float*** ImgMod,int lgth2,int wdth2)
{
 int k;

  //AllocationMemoire
  float** HistMod=fmatrix_allocate_2d(TROIS,NBBIN_IMG);
  float*** ImgLAB=fmatrix_allocate_3d(TROIS,lgth,wdth);
  float*** ImgLAB2=fmatrix_allocate_3d(TROIS,lgth,wdth);
  float*** ModLAB=fmatrix_allocate_3d(TROIS,lgth2,wdth2);

  //Conversion RGB>LAB
  ConvertImgRGBToLAB(ImgIn,ImgLAB,lgth,wdth);
  ConvertImgRGBToLAB(ImgMod,ModLAB,lgth2,wdth2);

  //ExactSpecif L/A/B
  for(k=0;k<TROIS;k++)
    { printf("  [%d]",k); 
      fflush(stdout);
      SPEC_ComputeHisto(ModLAB[k],HistMod[k],lgth2,wdth2);
      SPEC_SpecifHistoFromDist(ImgLAB[k],ImgLAB2[k],lgth,wdth,HistMod[k]); }

  //Conversion LAB>RGB
  ConvertImgLABToRGB(ImgLAB2,ImgOut,lgth,wdth);
  
  //Liberation Memoire
  if (HistMod)  free_fmatrix_2d(HistMod);
  if (ImgLAB)   free_fmatrix_3d(ImgLAB,TROIS);
  if (ImgLAB2)  free_fmatrix_3d(ImgLAB2,TROIS);
  if (ModLAB)   free_fmatrix_3d(ModLAB,TROIS);
} 

//----------------------------------------------------------
//  SPEC >> SpecifHistoFromImgNP
//----------------------------------------------------------
//   
//    Nota::Specif HistoCoul 
//
// Methode::
//       
//           -1- Suivant la relation d'ordre sur Lum(ImgIn)
//                  et celle de Lum(ImgTarget)
//           -2- Affectation Pixel_RVB_target sur Pixel_RVB_input
//       
//        Img in [0.0::1.0] ou [0::256]
//        Histo[NBBIN_IMG]
// 
//----------------------------------------------------------     
void EdgeSpecif::SPEC_SpecifHistoFromImgNP(float*** ImgIn,float*** ImgOut,int lgth,int wdth,
float*** ImgMod,int lgth2,int wdth2)
{
 int i,j,k;
 int nbp1,nbp2;
 int NbPixs1,NbPixs2;
 int posr1,posc1;
 int posr2,posc2;
 float MinDiff;
 float fact;

 //Constantes
 NbPixs1=lgth*wdth;
 NbPixs2=lgth2*wdth2;

 //AllocationMemoire
 float** TabHistImg=fmatrix_allocate_2d(NbPixs1,TROIS);
 float** TabHistMod=fmatrix_allocate_2d(NbPixs2,TROIS);
 float** LumImg=fmatrix_allocate_2d(lgth,wdth);
 float** LumMod=fmatrix_allocate_2d(lgth2,wdth2);

 //>ComputeLuminance
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
   LumImg[i][j]=(0.3*ImgIn[0][i][j])+(0.59*ImgIn[1][i][j]+0.11*ImgIn[2][i][j]); 

 for(i=0;i<lgth2;i++) for(j=0;j<wdth2;j++) 
   LumMod[i][j]=(0.3*ImgMod[0][i][j])+(0.59*ImgMod[1][i][j]+0.11*ImgMod[2][i][j]); 

 //>Trie Img
 MinDiff=10000;
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)
   { nbp1=(i*wdth)+j;
     TabHistImg[nbp1][0]=LumImg[i][j];
     TabHistImg[nbp1][1]=i;
     TabHistImg[nbp1][2]=j; 
     if ((LumImg[i][j])&&(LumImg[i][j]<MinDiff)) MinDiff=LumImg[i][j]; }

 printf("\n  > MinDiff=%.10f",MinDiff);
 for(k=0;k<NbPixs1;k++) TabHistImg[k][0]+=(randomize()*(MinDiff/100.0));  
 printf("\n  > Sort ..."); fflush(stdout);
 sort(TabHistImg,0,NbPixs1,TROIS);
 printf("..."); fflush(stdout);

 //>Trie Mod
 MinDiff=10000;
 for(i=0;i<lgth2;i++) for(j=0;j<wdth2;j++)
   { nbp2=(i*wdth2)+j;
     TabHistMod[nbp2][0]=LumMod[i][j];
     TabHistMod[nbp2][1]=i;
     TabHistMod[nbp2][2]=j; 
     if ((LumMod[i][j])&&(LumMod[i][j]<MinDiff)) MinDiff=LumMod[i][j]; }

 printf("\n  > MinDiff=%.10f",MinDiff);
 for(k=0;k<NbPixs2;k++) TabHistMod[k][0]+=(randomize()*(MinDiff/100.0));  
 printf("\n  > Sort ..."); fflush(stdout);
 sort(TabHistMod,0,NbPixs2,TROIS);
 printf("..."); fflush(stdout);

 //>Facteur
 fact=(float)NbPixs1/(float)NbPixs2;

 //>SPECIFICATION
 //============================================
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)
   { nbp1=(i*wdth)+j;
     nbp2=(int)((float)nbp1/fact);
     if (nbp2>(lgth2*wdth2)) nbp2=(lgth2*wdth2)-1;
     posr1=(int)TabHistImg[nbp1][1];
     posc1=(int)TabHistImg[nbp1][2];
     posr2=(int)TabHistMod[nbp2][1];
     posc2=(int)TabHistMod[nbp2][2];
     ImgOut[0][posr1][posc1]=ImgMod[0][posr2][posc2];
     ImgOut[1][posr1][posc1]=ImgMod[1][posr2][posc2];
     ImgOut[2][posr1][posc1]=ImgMod[2][posr2][posc2]; } 
 //============================================

 //LiberationMemoire
 if (TabHistImg) free_fmatrix_2d(TabHistImg);
 if (TabHistMod) free_fmatrix_2d(TabHistMod); 
 if (LumImg)     free_fmatrix_2d(LumImg); 
 if (LumMod)     free_fmatrix_2d(LumMod); 
} 

//==================================================
//==================================================
//  SPECIFICATION HISTOGRAMME GRADIENT LUMINANCE
//==================================================
//==================================================
//----------------------------------------------------------
// SPEC >> SpecifGradHistoFromMod
//----------------------------------------------------------
//
// SpecifGradHistoFrom[1]Mod 
// float** Img    :: Luminance Img
// float** ImgMod :: Img Modele
// float** ImgOut :: ImgResultat
//
//----------------------------------------------------------    
void EdgeSpecif::SPEC_SpecifGradHistoFromMod(float** Img,int lgth,int wdth,
int SzW,int NbCnx,float** ImgOut,int ImgMod,int NbBins,int FlagS)
{
 int k;
 int NbConnex;
 int Nghb;
 float tmp;

 //>Parametre
 NbConnex=NbCnx;
 Nghb=(int)(SzW/2);

  //>Presentation
  printf("\n\n ______________________________________");
  printf("\n ____SPEC SpecifGradHistoFrom[1]Mod____");
  printf("\n ______________________________________");
  printf("\n\n  >> Img   [Lgth=%d::Wdth=%d]",lgth,wdth);
  printf(" :: NbCnx [%d]",NbConnex);
  printf("\n  >> StepGradient [%.2f]",StGr);
  printf("\n  >> NbIterMax    [%d][%d]",NBITERMAX_GRAD,NBITERMAX_METR);
  printf("\n  -----------------");
  printf("\n  >> Dist. Law :: [%d] (0>Uni 1>SemiGaus  2>Trg  3>Shifted Gaus 4> Exp 5-8-> Edge Class)",ImgMod);
  printf("\n  -----------------");
  fflush(stdout);

  //>AllocationMemoire 
  float**  NonLocGrphDif=fmatrix_allocate_2d(NbConnex,(wdth*lgth));
  float**  ImgInit=fmatrix_allocate_2d(length,width);
  float*   HistGradMod=fmatrix_allocate_1d(NbBins);
  float*   HistGradImg=fmatrix_allocate_1d(NbBins);

  //>Img>ImgInit
  CopyMat(ImgInit,Img,lgth,wdth);

  //ComputeHistoGradLuminanceModel
  for(k=0;k<NbBins;k++) HistGradMod[k]=SPEC_HistoModels((float)k/NbBins,0.05,ImgMod);
  for(tmp=0.0,k=0;k<NbBins;k++) tmp+=HistGradMod[k];
  for(k=0;k<NbBins;k++) HistGradMod[k]/=tmp;

  //>ComputeBhattacharya Distance <Before> 
  SPEC_ComputeGradHisto(Img,HistGradImg,1,lgth,wdth);
  if (!Flag_SPEC) BhatBef=SPEC_ComputeBhattaDist(HistGradMod,HistGradImg,NbBins);
  printf("\n  >> DistanceBhatta<Before> = [%.4f]",BhatBef);
  
  //>NonLocalMultiResGraph
  SPEC_ComputeNonLocalGraphWindow(Img,lgth,wdth,NonLocGrphDif,SzW,NbCnx,HistGradMod,NbBins);
 
  //>ConjugateGradientDecente ImgRed>ImgRes
  SPEC_ConjGradDesc(ImgInit,ImgOut,lgth,wdth,NonLocGrphDif,NbCnx,NBITERMAX_GRAD);
  if (FlagS) SPEC_SimulatedAnnealing(ImgOut,ImgOut,lgth,wdth,NonLocGrphDif,NbCnx,NBITERMAX_METR,TEMP_INIT,RADIUS_EXPLOR);
  RecalI(ImgOut,lgth,wdth,0.0,1.0);

  //>ComputeBhattacharya Distance <After>
  SPEC_ComputeGradHisto(ImgOut,HistGradImg,1,lgth,wdth);
  BhatAft=SPEC_ComputeBhattaDist(HistGradMod,HistGradImg,NbBins);
  printf("\n  >> DistanceBhatta<After> = [%.4f]",BhatAft);

 //Liberation Memoire
 if (NonLocGrphDif) free_fmatrix_2d(NonLocGrphDif);
 if (ImgInit)       free_fmatrix_2d(ImgInit);
 if (HistGradMod)   free_fmatrix_1d(HistGradMod);
 if (HistGradImg)   free_fmatrix_1d(HistGradImg);
}

//----------------------------------------------------------
// SPEC >> SpecifGradHistoFromImg
//----------------------------------------------------------
//
// SpecifGradHistoFromImg
// float** Img    :: Luminance Img
// float** ImgMod :: Img Modele
// float** ImgOut :: ImgResultat
//
//----------------------------------------------------------    
void EdgeSpecif::SPEC_SpecifGradHistoFromImg(float** Img,int lgth,int wdth,
int SzW,int NbCnx,float** ImgOut,float** ImgMod,int lgth2,int wdth2,int NbBins)
{
 int k;
 int NbConnex;
 int Nghb;
 float tmp;

 //>Parametre
 NbConnex=NbCnx;
 Nghb=(int)(SzW/2);

  //>Presentation
  printf("\n\n ___________________________________");
  printf("\n ____SPEC SpecifGradHistoFromImg____");
  printf("\n ___________________________________");
  printf("\n\n  >> Img   [Lgth=%d::Wdth=%d]",lgth,wdth);
  printf(" :: NbCnx      [%d]",NbConnex);
  printf(" :: Nghb Order [%d]",Nghb);
  printf("\n  -----");
  printf("\n  >> StepGradient [%.2f]",StGr);
  printf("\n  >> NbIterMax    [%d][%d]",NBITERMAX_GRAD,NBITERMAX_METR);
  fflush(stdout);

  //>AllocationMemoire 
  float**  NonLocGrphDif=fmatrix_allocate_2d(NbConnex,(wdth*lgth));
  float**  ImgInit=fmatrix_allocate_2d(length,width);
  float**  HistGradMod=fmatrix_allocate_2d(Nghb,NbBins);
  float*   HistGradImg=fmatrix_allocate_1d(NbBins);

  //>Img>ImgInit
  CopyMat(ImgInit,Img,lgth,wdth);

  //ComputeHistoGradLuminanceModel
  for(k=0;k<Nghb;k++) 
    { printf("\n  >> SPEC_ComputeGradHisto[%d]",k);
      fflush(stdout);
      SPEC_ComputeGradHisto(ImgMod,k+1,HistGradMod[k],lgth2,wdth2); }

  //>ComputeBhattacharya Distance <Before> 
  for(k=0;k<Nghb;k++)
     { SPEC_ComputeGradHisto(Img,HistGradImg,k+1,lgth,wdth);
       tmp=SPEC_ComputeBhattaDist(HistGradMod[k],HistGradImg,NbBins);
       printf("\n  >> DistanceBhatta[%d]<Before> = [%.4f]",k,tmp); 
       if ((!Flag_SPEC)&&(!k)) BhatBef=tmp; }
  
  //>NonLocalMultiResGraph
  SPEC_ComputeNonLocalGraphWindow(Img,lgth,wdth,NonLocGrphDif,SzW,NbCnx,HistGradMod,NbBins);
 
  //>ConjugateGradientDecente ImgRed>ImgRes
  SPEC_ConjGradDesc(ImgInit,ImgOut,lgth,wdth,NonLocGrphDif,NbCnx,NBITERMAX_GRAD);  
  SPEC_SimulatedAnnealing(ImgOut,ImgOut,lgth,wdth,NonLocGrphDif,NbCnx,NBITERMAX_METR,TEMP_INIT,RADIUS_EXPLOR);
  RecalI(ImgOut,lgth,wdth,0.0,1.0);

  //>ComputeBhattacharya Distance <After>
  for(k=0;k<Nghb;k++)
     { SPEC_ComputeGradHisto(ImgOut,HistGradImg,k+1,lgth,wdth);
       tmp=SPEC_ComputeBhattaDist(HistGradMod[k],HistGradImg,NbBins);
       printf("\n DistanceBhatta[%d]<After> = [%.4f]",k,tmp); 
       if (!k) BhatAft=tmp; }

 //Liberation Memoire
 if (NonLocGrphDif) free_fmatrix_2d(NonLocGrphDif);
 if (ImgInit)       free_fmatrix_2d(ImgInit);
 if (HistGradMod)   free_fmatrix_2d(HistGradMod); 
 if (HistGradImg)   free_fmatrix_1d(HistGradImg);
}


//==================================================
//==================================================
//  SPECIFICATION-OPTIMISATION
//==================================================
//==================================================
//----------------------------------------------------------
// SPEC >> ComputeNonLocalGraphWindow
//----------------------------------------------------------
//
// [SIMPLE EDGE HISTOGRAMME]
//
// float** Img   :: Luminance 
// int lgth/wdth :: Longueur/Largeur originale
// HistGradMod   :: GradHisto Modele[NBIN_IMG]
//
//---------------------------------------------------------- 
void EdgeSpecif::SPEC_ComputeNonLocalGraphWindow(float** Img,int lgth,int wdth,\
float** NonLocGrphDif,int SzW,int NbCnx,float* HistGradMod,int NbBins)
{
 short int i,j,k,l;
 short int posc,posr;
 short int SzWdiv2;
 int cpt,nb;
 float ValMaxGrph;

 //Paramters
 SzWdiv2=(SzW/2);

 //Presentation
 printf("\n");  
 printf("\n  >> Compute SPEC__NonLocalGraph Window");
 printf("\n  >> Img:[lgth:%d-wdth:%d]",lgth,wdth);
 printf("  | Window Sz :[%d]",SzW);
 printf("  |  Connections :[%d]",NbCnx);
 fflush(stdout);

 //Init_NonLocGrph 
 for(i=0;i<lgth;i++) 
     for(j=0;j<wdth;j++) 
         for(k=0;k<NbCnx;k++) NonLocGrphDif[k][(i*wdth)+j]=0.0; 
     
  //-----------------------------------------------
  //--NonLocal Graph ------------------------------
  //-----------------------------------------------
  ValMaxGrph=0;

  for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)
     {
      //>SearchInBigWindow[posr,posc]
      cpt=0; nb=0; 
      for(k=0;k<SzW;k++) for(l=0;l<SzW;l++) 
        { 
          cpt++;
          posr=i-(SzWdiv2)+k;
          posc=j-(SzWdiv2)+l;
          if (posr<0)         posr+=lgth;
          if (posr>(lgth-1))  posr-=lgth; 
          if (posc<0)         posc+=wdth;
          if (posc>(wdth-1))  posc-=wdth;
       
          //>Beta::Abs(FirstOrderDiff)
          //--------------------------
          NonLocGrphDif[nb][(i*wdth)+j]=fabs(Img[i][j]-Img[posr][posc]);
                                 
          //>Incremente
	  nb++; 
	}   
     }//------------------------------------------

 //Specification Histo
 SPEC_SpecHistoGradOnModel(NonLocGrphDif,NbCnx,lgth,wdth,HistGradMod,NbBins);
}

//----------------------------------------------------------
// SPEC >> ComputeNonLocalGraphWindow
//----------------------------------------------------------
//
// [MULTIPLE EDGE HISTOGRAMS]
//
// float** Img   :: Luminance 
// int lgth/wdth :: Longueur/Largeur originale
// HistGradMod   :: GradHisto Modele[Nghb][BBIN_IMG]
//                      Nghb=SzWdiv2       
//
//---------------------------------------------------------- 
void EdgeSpecif::SPEC_ComputeNonLocalGraphWindow(float** Img,int lgth,int wdth,\
float** NonLocGrphDif,int SzW,int NbCnx,float** HistGradMod,int NbBins)
{
 short int i,j,k,l;
 short int posc,posr;
 short int SzWdiv2;
 int cpt,nb;
 float ValMaxGrph;

 //Paramters
 SzWdiv2=(SzW/2);

 //Presentation
 printf("\n");  
 printf("\n  >> Compute SPEC__NonLocalGraph Window");
 printf("\n  >> Img:[lgth:%d-wdth:%d]",lgth,wdth);
 printf("\n  >> NbModels:[%d]",SzWdiv2);
 printf("  | Window Sz :[%d]",SzW);
 printf("  |  Connections :[%d]",NbCnx);
 fflush(stdout);

 //Init_NonLocGrph 
 for(i=0;i<lgth;i++) 
     for(j=0;j<wdth;j++) 
         for(k=0;k<NbCnx;k++) NonLocGrphDif[k][(i*wdth)+j]=0.0; 
     
  //-----------------------------------------------
  //--NonLocal Graph ------------------------------
  //-----------------------------------------------
  ValMaxGrph=0;

  for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)
     {
      //>SearchInBigWindow[posr,posc]
      cpt=0; nb=0; 
      for(k=0;k<SzW;k++) for(l=0;l<SzW;l++) 
        { 
          cpt++;
          posr=i-(SzWdiv2)+k;
          posc=j-(SzWdiv2)+l;
          if (posr<0)         posr+=lgth;
          if (posr>(lgth-1))  posr-=lgth; 
          if (posc<0)         posc+=wdth;
          if (posc>(wdth-1))  posc-=wdth;
 
          //>Beta::Abs(FirstOrderDiff)
          //--------------------------
          NonLocGrphDif[nb][(i*wdth)+j]=fabs(Img[i][j]-Img[posr][posc]);
                                 
          //>Incremente
	  nb++; 
	}   
     }//------------------------------------------

 //Specification Histo
 SPEC_SpecHistoGradOnModel(NonLocGrphDif,NbCnx,lgth,wdth,HistGradMod,NbBins);
}

//----------------------------------------------------------
// SPEC >> ConjGradDesc 
//----------------------------------------------------------
//
// x[n+1]=x[n]+StGr*d[n]
// 
//  d[n]=- Grad_J(x[n])
//       + <Grad_J(x[n])-Grad_J(x[n-1]),Grad_J(x[n])>/<Grad_J(x[n-1]),Grad_J(x[n-1])> d[n-1]
//
// Img,Img7,Img8 : Img Initiale :: Image_Nrj_init
// Img1          : x[n] puis ...x[n+1]
// Img2          : Grad_J(x[n])
// Img3          : Grad_J(x[n-1])
// Img4          : d[n]
// Img5          : d[n-1]
// Img9          : x[n+1]
//----------------------------------------------------------    
void EdgeSpecif::SPEC_ConjGradDesc(float** Img,float** ImgRes,int lgth,int wdth,\
float** NonLocGrphDif,int NbCnx,int NbIter)
{
 int k;
 int NbDiv;
 float nrj,nrj_prec,nrj_init,nrj_best;
 float Beta;

 //>Init
 
 CopyMat(Img1,Img,lgth,wdth);
 CopyMat(Img8,Img,lgth,wdth);
 CopyMat(Img7,Img,lgth,wdth);
 nrj=SPEC_ComputeNrj(Img8,lgth,wdth,NonLocGrphDif,NbCnx);
 nrj_prec=nrj;
 nrj_init=0.0;
 nrj_init=nrj;
 printf("\n\n    >  ConjGradDescente [NbIterMax=%d][StepGr=%.2f]",NbIter,StGr);  
 printf("[MoyImg=%.4f]",ComputeMean(Img,lgth,wdth));  
 printf("\n\n    >  <NrjInit=%.6f>",nrj);
 fflush(stdout);
 if (!Flag_SPEC) NrjImgInit=nrj_init;
 nrj_best=nrj_init;
 Beta=0.0;
 NbDiv=0;

 //>Flag_stop
 int FlagStop=0; 

 //>Conjugate Gradient Descent
 //===========================
 for(k=0;((k<=NbIter)&&(!FlagStop));k++)
   {
     //>Info
     printf("\n [%d]>  ",k); 
    
     // Img2 <-- Grad_J(x[k]) 
     SPEC_DeriveNrj(Img1,Img2,lgth,wdth,NonLocGrphDif,NbCnx);

     //-------------------------------------------
     // [k==0] SimpleGradient
     //-------------------------------------------
     //      > d[k] <-- Grad_J(x[k])    
     //      > x[k+1]=x[n]-StrGr*Grad_J(x[k]) 
     if (k==0) 
        { CopyMat(Img4,Img2,lgth,wdth);
          AddMatrix(Img9,Img1,-1.0*StGr,Img2,lgth,wdth); }

     //--------------------------------------------
     // [k!=0] Gradient Conjugue
     //-------------------------------------------- 
     //      > x[k+1]=x[k]-StGr*d[k]
     if (k) 
       {  Beta=ComputeBeta(Img2,Img3,lgth,wdth);
          MultMatrix(Img6,Beta,Img5,lgth,wdth);
          AddMatrix(Img4,Img2,-1.0,Img6,lgth,wdth);
          AddMatrix(Img9,Img1,-1.0*StGr,Img4,lgth,wdth); }

     //>Calcul Nrj 
     nrj_prec=nrj;
     nrj=SPEC_ComputeNrj(Img9,lgth,wdth,NonLocGrphDif,NbCnx);
     printf("<Nrj=%.6f>",nrj);
     fflush(stdout);

    //>Record ImgBest>Img8 & NrjBest
    if ((nrj<nrj_prec)&&(nrj<nrj_best)) CopyMat(Img8,Img9,lgth,wdth);  
    if (nrj<nrj_best) nrj_best=nrj;
 
    //>STOP CRITERION
    //--------------------------------------

    //1--> Nrj Increases 
    if  ((nrj>nrj_prec)||(isnan(nrj))||(nrj>nrj_best))
	 { CopyMat(Img9,Img8,lgth,wdth);
           StGr/=2.0;
           InitMatrix(Img5,lgth,wdth);
           NbDiv++;
           printf(" %d>NewStep=%.6f",NbDiv,StGr);
           k--;
           fflush(stdout); }
    //-------------------------------------  

    // x[k+1] --> x[k]  
    // Img3 <-- Grad_J(x[k-1]) 
    // Img5 <-- d[k-1]
    CopyMat(Img1,Img9,lgth,wdth);
    CopyMat(Img3,Img2,lgth,wdth);
    CopyMat(Img5,Img4,lgth,wdth);

    //End
    if ((NbDiv>4)||(k>NbIter)) break;  

   }//FIN_ALGO======================================

 //Enregistrement Nrj_finale[Img9] & Stabilite[Img9]
 NrjImg=nrj;
 NbIterFinal=k;

 printf("\n   >> [Nrj:%.7f]",NrjImg);
 printf("        [MoyImg=%.4f]",ComputeMean(Img9,lgth,wdth)); 
 printf("[DiffImg=%.4f]\n",ComputeDiff(Img9,Img7,lgth,wdth));  
 fflush(stdout);

 //Transfert>ImgRes
 CopyMat(ImgRes,Img9,lgth,wdth);
}

//----------------------------------------------------------
// SPEC >> DeriveNrj
//----------------------------------------------------------
//
// DERIV= 4 SUM_<s,t> (Xs -Xt) [ Beta^2-(Xs-Xt)^2 ]  
//
//----------------------------------------------------------  
void EdgeSpecif::SPEC_DeriveNrj(float** Img,float** ImgGrad,int lgth,int wdth,\
float** NonLocGrphDif,int NbCnx)
{
 int i,j,k;
 int posr,posc;
 float tmp;
 float xs,xt;
 float difx;
 int nbp;
 int SzW;

 //Init
 SzW=(int)sqrt(NbCnx);
 
 //Boucle
 //====================================
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)
   {
     nbp=(i*wdth)+j;
     xs=Img[i][j];

     tmp=0.0;
     //>ForAllConnexion
     for(k=0;k<NbCnx;k++)
       { 
         SPEC_GivePosPixel(i,j,k,&posr,&posc,SzW,lgth,wdth);
         xt=Img[posr][posc];
         difx=(xs-xt); 
         tmp+=-4*(CARRE(NonLocGrphDif[k][nbp])-CARRE(difx))*difx;
        }

     //>Record
     ImgGrad[i][j]=tmp/(float)NbCnx;
   } 
}

//----------------------------------------------------------
// SPEC >> Compute NRJ  
//----------------------------------------------------------
//
// NRJ = SUM_<st> ( Beta^2 - (Xs-Xt)^2 )^2  [Ok]
//---------------------------------------------------------- 
float EdgeSpecif::SPEC_ComputeNrj(float** Img,int lgth,int wdth,float** NonLocGrphDif,int NbCnx)
{
 int i,j,k;
 float totvr,TotalVar;
 int posr,posc;
 int nbp;
 float dist;
 int SzW;

 //Init
 SzW=(int)sqrt(NbCnx);

 //Calcul NRJ
 //=====================================
 TotalVar=0.0;
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
     {
      totvr=0.0;
      for(k=0;k<NbCnx;k++)
        { nbp=(i*wdth)+j; //A optimiser
          SPEC_GivePosPixel(i,j,k,&posr,&posc,SzW,lgth,wdth);
          dist=NonLocGrphDif[k][nbp];
          totvr+=CARRE(CARRE(dist)-CARRE(Img[i][j]-Img[posr][posc])); }

      TotalVar+=totvr/(float)NbCnx;
     }

 //Retour
 return TotalVar;
}

//----------------------------------------------------------
// SPEC >> GivePosPixel  
//----------------------------------------------------------
//
//----------------------------------------------------------
void EdgeSpecif::SPEC_GivePosPixel(int posr,int posc,int NbC,int* row,int* col,int SzW,int lgth,int wdth)
{
 int RowInWnd,ColInWnd;
 int SzWd2;

 SzWd2=(int)(SzW/2);
 RowInWnd=(NbC/SzW)+posr-SzWd2; 
 ColInWnd=(NbC%SzW)+posc-SzWd2; 
 if (RowInWnd<0)         RowInWnd+=lgth;
 if (RowInWnd>(lgth-1))  RowInWnd-=lgth; 
 if (ColInWnd<0)         ColInWnd+=wdth;
 if (ColInWnd>(wdth-1))  ColInWnd-=wdth;
 (*row)=RowInWnd;
 (*col)=ColInWnd;
}

//---------------------------------------------------------------------
//  SPEC >> SpecHistoGradOnModel
//---------------------------------------------------------------------
//   
//    MODEL :: 
//          :: k \in ]0.0::1.0] en NbBins bins
//
//   HistoGrad :: \in [0.0::1.0] en NbBind bins
//             :: suivant MODEL DISTRIBUTION
//----------------------------
//
//  TabHistGrad[l][0] :: HistGrad. Val. In < l \in [0::lgth*wdth*(NbCnx-1)] >
//             [l][1] :: (i*wdth)+j
//             [l][2] :: k= 0:: NbCx-1
//             [l][3] :: HistGrad. Val. Out
//      
//      HistGradMod   :: GradHisto Modele[NbBins]
//---------------------------------------------------------------------     
void EdgeSpecif::SPEC_SpecHistoGradOnModel(float** NLGrph,int NbCx,int lgth,int wdth,float* HistoGrad,int NbBins)
{
 int i,j,k,l,m,n;
 int nbc,nbp,nbdf;
 int NbDiff;
 int NbCols;
 int pos;
 float MinDiff;
 float PropPixs;
 float idxmod;
 float tmp;

 int NbCxMax;
 int NbVois;
 int SzW;
 int SzWd2;
 float dist;

 //Calcul TabPositions
 SzW=(int)sqrt(NbCx);
 SzWd2=(int)(sqrt(NbCx)/2.0);
 NbVois=(int)(sqrt(NbCx)/2.0);
 NbCxMax=8*NbVois;
 
 //Constantes
 NbCols=4;

 //Presentation
 printf("\n\n  >> SPECIFICATION [NbVois::%d][NbBins::%d]",NbVois,NbBins);
 printf("\n");
 fflush(stdout);

 //AllocationMemoire
 int** TabPos=imatrix_allocate_2d(NbVois,NbCxMax);
 float** TabHistGrad=fmatrix_allocate_2d(lgth*wdth*NbCxMax,NbCols);
 float** HistModel=fmatrix_allocate_2d(2,NbBins);

 //Initialisation
 for(k=0;k<NbVois;k++) for(l=0;l<NbCxMax;l++) TabPos[k][l]=-1;
 for(k=0;k<lgth*wdth*NbCxMax;k++) for(l=0;l<NbCols;l++) TabHistGrad[k][l]=0.0;
 for(l=0;l<2;l++) for(k=0;k<NbBins;k++) HistModel[l][k]=0.0;
 
 //Remplissage TabPos
 for(nbp=0,m=-SzWd2;m<=SzWd2;m++) for(n=-SzWd2;n<=SzWd2;n++)
    { dist=(int)MAX(fabs(m),fabs(n))-1;
      if (dist<0) { nbp++; continue; }
      for(l=0;TabPos[(int)dist][l]>=0;l++)  continue; 
      TabPos[(int)dist][l]=nbp++; }

 //DEBUG
 if (0) { printf("\n");
          for(k=0;k<NbVois;k++) 
             { printf("\n");
               for(l=0;l<NbCxMax;l++)  printf("[%d]",TabPos[k][l]); }
               exit(-1); }

  //==================
  //BOUCLE
  //==================
  for(n=0;n<NbVois;n++)
    {
     printf("\n   [%d] Specification...",n+1);
     fflush(stdout);
     NbDiff=lgth*wdth*(8*(n+1));

           //>RecupVal & FindMinDiff
           MinDiff=10000;
           nbdf=0;
           for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)
              { 
               nbp=(i*wdth)+j;
               for(k=0;k<(8*(n+1));k++) 
                  { 
		   pos=TabPos[n][k];
                
                   TabHistGrad[nbdf][0]=NLGrph[pos][nbp]; 
                   TabHistGrad[nbdf][1]=nbp;
                   TabHistGrad[nbdf][2]=pos;
                   nbdf++; 
                   if ((NLGrph[pos][nbp])&&(NLGrph[pos][nbp]<MinDiff)) { MinDiff=NLGrph[pos][nbp]; } 
                  }      
	      }
           printf("  > MinDiff=%.10f",MinDiff);
           fflush(stdout);

           //>Add MinDiff/4 Noise 
           for(k=0;k<NbDiff;k++) TabHistGrad[k][0]+=(randomize()*(MinDiff/4.0));

           //>Sort
           printf("  > Sort"); fflush(stdout);
           sort(TabHistGrad,0,NbDiff,NbCols);
           printf("..."); fflush(stdout);

          //>DEBUG
          if (0) for(k=0;k<NbDiff;k++)
                 { printf("\n");
                   for(l=0;l<NbCols;l++) printf("[%f]",TabHistGrad[k][l]); }

          //===============
          //>SPECIFICATION
          //=============== 
          for(k=0;k<NbBins;k++) HistModel[0][k]=HistoGrad[k];
          for(tmp=0.0,k=0;k<NbBins;k++) tmp+=HistModel[0][k];
          for(k=0;k<NbBins;k++) HistModel[0][k]/=tmp;

          HistModel[1][0]=HistModel[0][0];
          for(k=1;k<NbBins;k++) HistModel[1][k]=HistModel[1][k-1]+HistModel[0][k];
          printf(" > [Histo. Cum=%.2f]",HistModel[1][NbBins-1]);

          for(l=0,nbdf=0,k=0;k<NbBins;k++)
             { idxmod=(float)k/(float)NbBins;
               PropPixs=(int)((NbDiff*HistModel[1][k]));    
               for(l=nbdf;(l<(PropPixs))&&(l<NbDiff);l++) TabHistGrad[l][3]=idxmod;    
               nbdf=(int)PropPixs; }
 
          printf(" > [GradHisto::%d>%d]",nbdf,NbDiff); fflush(stdout);
          for(l=nbdf;l<NbDiff;l++) if (!TabHistGrad[l][3]) TabHistGrad[l][3]=TabHistGrad[l-1][3];
          //=============================================

          //>DEBUG
          if (0) for(k=0;k<NbDiff;k++)
                 { printf("\n");
                   for(l=0;l<NbCols;l++) printf("[%f]",TabHistGrad[k][l]); }

          //============================================
          //>TRANSFERT>NLGrph
          //
          // TabHistGrad[l][1] :: (i*wdth)+j
          //            [l][2] :: k= 0:: NbCx-1
          //            [l][3] :: HistGrad. Val. Out
          //============================================
          for(l=0;l<NbDiff;l++)
             { nbp=(int)TabHistGrad[l][1];
               nbc=(int)TabHistGrad[l][2];  
               idxmod=TabHistGrad[l][3];
               NLGrph[nbc][nbp]=idxmod; }
          //============================================

    }//==========================================================================

 //LiberationMemoire
 if (TabHistGrad) free_fmatrix_2d(TabHistGrad);
 if (HistModel)   free_fmatrix_2d(HistModel); 
 if (TabPos)      free_imatrix_2d(TabPos);
}

//---------------------------------------------------------------------
//  SPEC >> SpecHistoGradOnModel
//---------------------------------------------------------------------
//   
//    MODEL :: 
//          :: k \in ]0.0::1.0] en NbBins bins
//
//   HistoGrad :: \in [0.0::1.0] en NbBind bins
//             :: suivant MODEL DISTRIBUTION
//----------------------------
//
//  TabHistGrad[l][0] :: HistGrad. Val. In < l \in [0::lgth*wdth*(NbCnx-1)] >
//             [l][1] :: (i*wdth)+j
//             [l][2] :: k= 0:: NbCx-1
//             [l][3] :: HistGrad. Val. Out
//      
//      HistGradMod   :: GradHisto Modele[NbVois][NbBins]
//---------------------------------------------------------------------     
void EdgeSpecif::SPEC_SpecHistoGradOnModel(float** NLGrph,int NbCx,int lgth,int wdth,float** HistoGrad,int NbBins)
{
 int i,j,k,l,m,n;
 int nbc,nbp,nbdf;
 int NbDiff;
 int NbCols;
 int pos;
 float MinDiff;
 float PropPixs;
 float idxmod;
 float tmp;

 int NbCxMax;
 int NbVois;
 int SzW;
 int SzWd2;
 float dist;

 //Calcul TabPositions
 SzW=(int)sqrt(NbCx);
 SzWd2=(int)(sqrt(NbCx)/2.0);
 NbVois=(int)(sqrt(NbCx)/2.0);
 NbCxMax=8*NbVois;
 
 //Constantes
 NbCols=4;

 //Presentation
 printf("\n\n  >> SPECIFICATION [NbVois::%d][NbModels::%d][NbBins::%d]",NbVois,SzWd2,NbBins);
 printf("\n");
 fflush(stdout);

 //AllocationMemoire
 int** TabPos=imatrix_allocate_2d(NbVois,NbCxMax);
 float** TabHistGrad=fmatrix_allocate_2d(lgth*wdth*NbCxMax,NbCols);
 float** HistModel=fmatrix_allocate_2d(2,NbBins);

 //Initialisation
 for(k=0;k<NbVois;k++) for(l=0;l<NbCxMax;l++) TabPos[k][l]=-1;
 for(k=0;k<lgth*wdth*NbCxMax;k++) for(l=0;l<NbCols;l++) TabHistGrad[k][l]=0.0;
 for(l=0;l<2;l++) for(k=0;k<NbBins;k++) HistModel[l][k]=0.0;
 
 //Remplissage TabPos
 for(nbp=0,m=-SzWd2;m<=SzWd2;m++) for(n=-SzWd2;n<=SzWd2;n++)
    { dist=(int)MAX(fabs(m),fabs(n))-1;
      if (dist<0) { nbp++; continue; }
      for(l=0;TabPos[(int)dist][l]>=0;l++)  continue; 
      TabPos[(int)dist][l]=nbp++; }

 //DEBUG
 if (0) { printf("\n");
          for(k=0;k<NbVois;k++) 
             { printf("\n");
               for(l=0;l<NbCxMax;l++)  printf("[%d]",TabPos[k][l]); }
               exit(-1); }

  //==================
  //BOUCLE
  //==================
  for(n=0;n<NbVois;n++)
    {
     printf("\n   [%d] Specification...",n+1);
     fflush(stdout);
     NbDiff=lgth*wdth*(8*(n+1));

           //>RecupVal & FindMinDiff
           MinDiff=10000;
           nbdf=0;
           for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)
              { 
               nbp=(i*wdth)+j;
               for(k=0;k<(8*(n+1));k++) 
                  { 
		   pos=TabPos[n][k];
                
                   TabHistGrad[nbdf][0]=NLGrph[pos][nbp]; 
                   TabHistGrad[nbdf][1]=nbp;
                   TabHistGrad[nbdf][2]=pos;
                   nbdf++; 
                   if ((NLGrph[pos][nbp])&&(NLGrph[pos][nbp]<MinDiff)) { MinDiff=NLGrph[pos][nbp]; } 
                  }      
	      }
           printf("  > MinDiff=%.10f",MinDiff);
           fflush(stdout);

           //>Add MinDiff/4 Noise 
           for(k=0;k<NbDiff;k++) TabHistGrad[k][0]+=(randomize()*(MinDiff/4.0));

           //>Sort
           printf("  > Sort"); fflush(stdout);
           sort(TabHistGrad,0,NbDiff,NbCols);
           printf("..."); fflush(stdout);

          //>DEBUG
          if (0) for(k=0;k<NbDiff;k++)
                 { printf("\n");
                   for(l=0;l<NbCols;l++) printf("[%f]",TabHistGrad[k][l]); }

          //===============
          //>SPECIFICATION
          //=============== 
          for(k=0;k<NbBins;k++) HistModel[0][k]=HistoGrad[n][k];
          for(tmp=0.0,k=0;k<NbBins;k++) tmp+=HistModel[0][k];
          for(k=0;k<NbBins;k++) HistModel[0][k]/=tmp;

          HistModel[1][0]=HistModel[0][0];
          for(k=1;k<NbBins;k++) HistModel[1][k]=HistModel[1][k-1]+HistModel[0][k];
          printf(" > [Histo. Cum=%.2f]",HistModel[1][NbBins-1]);

          for(l=0,nbdf=0,k=0;k<NbBins;k++)
             { idxmod=(float)k/(float)NbBins;
               PropPixs=(int)((NbDiff*HistModel[1][k]));    
               for(l=nbdf;(l<(PropPixs))&&(l<NbDiff);l++) TabHistGrad[l][3]=idxmod;    
               nbdf=(int)PropPixs; }
 
          printf(" > [GradHisto::%d>%d]",nbdf,NbDiff); fflush(stdout);
          for(l=nbdf;l<NbDiff;l++) if (!TabHistGrad[l][3]) TabHistGrad[l][3]=TabHistGrad[l-1][3];
          //=============================================

          //>DEBUG
          if (0) for(k=0;k<NbDiff;k++)
                 { printf("\n");
                   for(l=0;l<NbCols;l++) printf("[%f]",TabHistGrad[k][l]); }

          //============================================
          //>TRANSFERT>NLGrph
          //
          // TabHistGrad[l][1] :: (i*wdth)+j
          //            [l][2] :: k= 0:: NbCx-1
          //            [l][3] :: HistGrad. Val. Out
          //============================================
          for(l=0;l<NbDiff;l++)
             { nbp=(int)TabHistGrad[l][1];
               nbc=(int)TabHistGrad[l][2];  
               idxmod=TabHistGrad[l][3];
               NLGrph[nbc][nbp]=idxmod; }
          //============================================

    }//==========================================================================

 //LiberationMemoire
 if (TabHistGrad) free_fmatrix_2d(TabHistGrad);
 if (HistModel)   free_fmatrix_2d(HistModel); 
 if (TabPos)      free_imatrix_2d(TabPos);
}

//=================================
//=================================
//  COLOR & LUMINANCE TREATMENT
//=================================
//=================================
//---------------------------------------------------------------------
//  SPEC >> ConvertImgRGBToLum
//---------------------------------------------------------------------
//   
//---------------------------------------------------------------------     
void EdgeSpecif::SPEC_ConvertImgRGBToLum(float*** ImgIn,float** Lum,int lgth,int wdth)
{ 
int i,j;

 //AllocationMemoire
 float*** ImgLAB=fmatrix_allocate_3d(TROIS,lgth,wdth);

 //Conversion RGB>LAB
 ConvertImgRGBToLAB(ImgIn,ImgLAB,lgth,wdth);

 //Recup Lum
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) Lum[i][j]=ImgLAB[0][i][j];
 
 //Liberation Memoire 
 if (ImgLAB)   free_fmatrix_3d(ImgLAB,TROIS);  
}

//--------------------------------------------------
// SPEC_ConvertImgRGBToLum
//--------------------------------------------------
void EdgeSpecif::SPEC_ConvertImgRGBToLum_(float*** RGB,float** L,int lgth,int wdth)
{
 int i,j;

 //Boucle
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)
   L[i][j]=(0.3*RGB[0][i][j])+(0.59*RGB[1][i][j]+0.11*RGB[2][i][j]);
}

//---------------------------------------------------------------------
//  SPEC >> ReFindColor
//---------------------------------------------------------------------
//   
// float** ImgInit :: R,V,B \in [0.0::256.0]
//---------------------------------------------------------------------     
void EdgeSpecif::SPEC_ReFindColor(float** Lum,float*** ImgInit,float*** ImgOut,int lgth,int wdth)
{
 int i,j;
 float moyA,moyB;

  //Presentation
  printf("\n\n  >>> ReFindColor\n");
  fflush(stdout);

  //AllocationMemoire
  float*** ImgLAB=fmatrix_allocate_3d(TROIS,lgth,wdth);
 
  //Conversion RGB>LAB
  ConvertImgRGBToLAB(ImgInit,ImgLAB,lgth,wdth);

  //Factor
  moyA=0.0; 
  for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) moyA+=ImgLAB[0][i][j];
  moyB=0.0;
  for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) moyB+=Lum[i][j];

  //Add L
  for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
    if (moyB) ImgLAB[0][i][j]=Lum[i][j]*sqrt(moyA/moyB);

  //Conversion LAB>RGB
  ConvertImgLABToRGB(ImgLAB,ImgOut,lgth,wdth);
  
  //Liberation Memoire 
  if (ImgLAB)   free_fmatrix_3d(ImgLAB,TROIS);  
}

//---------------------------------------------------------- 
// SPEC >> BuiltColorImg 
//----------------------------------------------------------
//
//  ImgInit  : R,V,B [0::256]
//  LumSpec  : Luminance specifi [0::256] 
//  ImgOut   : R,V,B [0::256]
//            
//----------------------------------------------------------
void EdgeSpecif::SPEC_BuiltColorImg(float*** ImgInit,float** LumSpec,int lgth,int wdth,float*** ImgOut,float sat)
{ 
 int i,j,k;
 float tmp;

 //Presentation
 printf("\n > ReFindColor");
 fflush(stdout);

 //AllocationMemoire
 float** ImgL=fmatrix_allocate_2d(lgth,wdth);

 //ComputeLuminance
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)
   ImgL[i][j]=(0.3*ImgInit[0][i][j])+(0.59*ImgInit[1][i][j]+0.11*ImgInit[2][i][j]); 

 //>Conversion
 //=================
 for(k=0;k<TROIS;k++)
     for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)
       { tmp=0.0;
         if (ImgL[i][j]) tmp=ImgInit[k][i][j]/(0.5+ImgL[i][j]);  
	 ImgOut[k][i][j]=powf(tmp,sat)*LumSpec[i][j]; }    

 //Liberation Memoire 
 if (ImgL)  free_fmatrix_2d(ImgL);
}

//---------------------------------------------------------- 
// SPEC >> RecalSameMoyLum
//----------------------------------------------------------
//            
//----------------------------------------------------------
void EdgeSpecif::SPEC_RecalSameMoyLum(float*** ImgIn,float** LumSpecIn,int lgth,int wdth)
{ 
 int i,j;
 float MoyLumIn;
 float MoyLumSpecIn;

 //AllocationMemoire
 float*** ImgL=fmatrix_allocate_3d(TROIS,lgth,wdth);

 //Conversion ImgIn RGB> ImgIn L(AB)
 ConvertImgRGBToLAB(ImgIn,ImgL,lgth,wdth);
 
 //Calcul MoyLum(ImgIn) 
 MoyLumIn=0.0;
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) MoyLumIn+=ImgL[0][i][j];
 MoyLumIn/=(wdth*lgth);
 printf("\n Moyenne_Luminance_In=%.6f",MoyLumIn);

 //Calcul MoyLum(LumSpecIn) 
 MoyLumSpecIn=0.0;
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) MoyLumSpecIn+=LumSpecIn[i][j];
 MoyLumSpecIn/=(wdth*lgth);
 printf("\n Moyenne_Luminance_Spec_In=%.6f",MoyLumSpecIn);

 //Recalage
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) LumSpecIn[i][j]+=(MoyLumIn-MoyLumSpecIn);
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
   { if (LumSpecIn[i][j]<0.0) LumSpecIn[i][j]=0.0;
     if (LumSpecIn[i][j]>1.0) LumSpecIn[i][j]=1.0; }

 MoyLumSpecIn=0.0;
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) MoyLumSpecIn+=LumSpecIn[i][j];
 MoyLumSpecIn/=(wdth*lgth);
 printf("\n Moyenne_Luminance_Spec_In=%.6f",MoyLumSpecIn);

 //Liberation Memoire 
 if (ImgL)  free_fmatrix_3d(ImgL,TROIS);
}



//=================================
//=================================
//  DISTANCE & CRITERE
//=================================
//=================================
//---------------------------------------------------------- 
// SPEC >> SPEC_ComputeBhattaDist
//----------------------------------------------------------
// 
//----------------------------------------------------------
float EdgeSpecif::SPEC_ComputeBhattaDist(float* HistA,float* HistB,int nbbins)
{
  int k;
  float tmp;

  //>Normalisation
  for(tmp=0.0,k=0;k<nbbins;k++) tmp+=HistA[k];
  for(k=0;k<nbbins;k++) if (tmp) HistA[k]/=tmp;

  for(tmp=0.0,k=0;k<nbbins;k++) tmp+=HistB[k];
  for(k=0;k<nbbins;k++) if (tmp) HistB[k]/=tmp;

  //>BhattaDistance
  for(tmp=0.0,k=0;k<nbbins;k++) tmp+=sqrt(HistA[k]*HistB[k]);
  if (tmp>1.0) tmp=1.0;
  tmp=sqrt(1-tmp);

  //>Retour
  return tmp;
} 

//=================================
//=================================
//  INFO
//=================================
//=================================
//----------------------------------------------------------
// SPEC_FillInfoProcess 
//----------------------------------------------------------    
void EdgeSpecif::SPEC_FillInfoProcess()
{
 char Buff[NCHAR];
 char NameImg1[10];
 char NameImg2[10]; 

 //>CopiageName
 strcpy(NameImg1,"_________");
 strcpy(NameImg2,"_________");
 strncpy(NameImg1,Name_Img1,8);
 strncpy(NameImg2,Name_Img2,8);

 //>NbVoisinage[Ok]
 sprintf(Buff,"_SPEC_");
 strcpy(InfoProcess,Buff);

 //>Name[Ok]
 sprintf(Buff,"%s",NameImg1);
 strcat(InfoProcess,Buff);
 sprintf(Buff,"%s",NameImg2);
 strcat(InfoProcess,Buff);

 //>Method[Ok]
 sprintf(Buff,"_Meth_%d",(int)Alpha);
 strcat(InfoProcess,Buff);

 //>Law[Ok]
 if (Alpha==1)
    { sprintf(Buff,"_Law_%d",LawNumber);
      strcat(InfoProcess,Buff); 
      sprintf(Buff,"_Order_1");
      strcat(InfoProcess,Buff);   }

 if (Alpha>1)
    { sprintf(Buff,"_Order_%d",LawNumber);
      strcat(InfoProcess,Buff); }

 //>Nrj[Ok]
 sprintf(Buff,"_Nrj_%.0f_%.0f",NrjImgInit*100,NrjImg*100);
 strcat(InfoProcess,Buff);

 //>BhattaDist[Ok]
 sprintf(Buff,"_Dist_%.3f_%.3f",BhatBef,BhatAft);
 strcat(InfoProcess,Buff);

 //>NbIter[Ok]
 sprintf(Buff,"_NbIter_%d",NbIter);
 strcat(InfoProcess,Buff);

 //>Time[Ok]
 sprintf(Buff,"_Sec_%d",TimeSec);
 strcat(InfoProcess,Buff);

 //>StepGradient[Ok]
 //sprintf(Buff,"_StG_%.1f",StGrInit);
 //strcat(InfoProcess,Buff);

 //>Extension[Ok]
 sprintf(Buff,"_%s",Extension);
 strcat(InfoProcess,Buff);

 //>Display
 printf("\n Extension -> %s",InfoProcess);
}


//================================================
//================================================
//  RECUIT SIMULE
//================================================
//================================================
//----------------------------------------------------------
// SPEC >> Compute Local NRJ  
//----------------------------------------------------------
//
// NRJ = SUM_<st> ( Beta^2 - (Xs-Xt)^2 )^2
//---------------------------------------------------------- 
float EdgeSpecif::SPEC_ComputeLocNrj(float** Img,int row,int col,float val,int lgth,int wdth,float** NonLocGrphDif,int SzW,int NbCnx)
{
 int k;
 float totvr;
 int posr,posc;
 int nbp;
 float dist;

 //Calcul Local NRJ
 //-------------------------
 totvr=0.0;
 nbp=(row*wdth)+col;
 
 for(k=0;k<NbCnx;k++)
    { SPEC_GivePosPixel(row,col,k,&posr,&posc,SzW,lgth,wdth);
      dist=NonLocGrphDif[k][nbp];
      totvr+=CARRE(CARRE(dist)-CARRE(val-Img[posr][posc])); }

 totvr/=(float)NbCnx;

 //Retour
 return totvr;
}

//----------------------------------------------------------
// calcule le nombre de differences a chaque iteration
//----------------------------------------------------------
float EdgeSpecif::SPEC_DiffMat(float** pmatl1,float** pmatl2,int lgth,int wdth,float precs)
{
 int i,j;
 float chg;

 //>Compte NbChgt
 chg=0.0;
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
   if (fabs(pmatl1[i][j]-pmatl2[i][j])>precs) chg++; 
 chg/=(lgth*wdth);
 chg*=100.0;
       
 return chg;
}


//---------------------------------------------------------
//----------------------------------------------------------
// SPEC >> SimulatedAnnealing
//----------------------------------------------------------
//
//----------------------------------------------------------    
void EdgeSpecif::SPEC_SimulatedAnnealing(float** Img,float** ImgRes,int lgth,int wdth,\
					   float** NonLocGrphDif,int NbCnx,int NbIter,float Temp_Init,float Radius)
{
 int i,j,k;
 float nrj,nrj2;
 float ValImg;
 float rd,rd2; 
 float DeltaNrj;
 float ProbAcc;
 float RadiusDiv2; 

 //>Simulated Annealing
 float Temp_Fina=TEMP_FINAL;
 float Temp,TempInv;
 int   Kmax=NbIter;

 //>Init
 RadiusDiv2=Radius/2.0;
 SzW=(int)sqrt(NbCnx);
 CopyMat(Img1,Img,lgth,wdth);
 CopyMat(Img2,Img,lgth,wdth);
 nrj=SPEC_ComputeNrj(Img1,lgth,wdth,NonLocGrphDif,NbCnx);
 printf("\n\n    >  Simulated Annealing [NbIterMax=%d][Temp_Init=%f][Radius=%.3f]",NbIter,Temp_Init,Radius);  
 printf("[MoyImg=%.4f]",ComputeMean(Img,lgth,wdth));  
 printf("\n\n    >  <NrjInit=%.6f>",nrj);
 fflush(stdout);


  //====================================================================
  //Simulated Annealing -->sur Img 
  //====================================================================
  for(k=0;;k++)
     {
      Temp=Temp_Init*powf((Temp_Fina/Temp_Init),((float)k/Kmax));
      TempInv=1.0/Temp;
      if (Temp<Temp_Fina) break;

      if (!(k%10)) CopyMat(Img2,Img1,lgth,wdth);

      //-----------------------------------------------------------
      //>Loop Pixels
      //------------------------------------------------------------
      for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)
        {
           //Gibbs Sampling Record
	   ValImg=Img1[i][j];
  	   rd=ValImg-RadiusDiv2+(Radius*randomize());
 
           if (rd<0.0) rd=0.0;
           if (rd>1.0) rd=1.0;
 
            nrj2=SPEC_ComputeLocNrj(Img1,i,j,rd,lgth,wdth,NonLocGrphDif,SzW,NbCnx);
            nrj=SPEC_ComputeLocNrj(Img1,i,j,ValImg,lgth,wdth,NonLocGrphDif,SzW,NbCnx);
            DeltaNrj=nrj2-nrj;

            if (DeltaNrj<0) Img1[i][j]=rd;

            else { ProbAcc=exp(-TempInv*DeltaNrj); 
                   rd2=randomize();
                   if (rd2<ProbAcc)  Img1[i][j]=rd; }    
           
         }//End_Loop_Pixel------------------------------------------  

      //>Info
      if (!(k%10))
	 { nrj=SPEC_ComputeNrj(Img1,lgth,wdth,NonLocGrphDif,NbCnx); 
                  { printf("\n  [%d]>%.4f [Temp:%.1f 10^-9]",k,nrj,Temp*1000000000);
                   } }

     }//End_Simulated_Annealing==================================================================
  
 //Nrj_finale[Img1] 
 CopyMat(ImgRes,Img1,lgth,wdth);
 nrj=SPEC_ComputeNrj(ImgRes,lgth,wdth,NonLocGrphDif,NbCnx);
 NrjImg=nrj;
 printf("\n   >> [Nrj:%.7f]",nrj);
 printf("        [MoyImg=%.4f]",ComputeMean(ImgRes,lgth,wdth));   
 fflush(stdout);
}
